##################################################################
# Copyright 2018 Open Source Geospatial Foundation and others    #
# licensed under MIT, Please consult LICENSE.txt for details     #
##################################################################

from json import JSONEncoder


class ArrayEncoder(JSONEncoder):
    def default(self, obj):
        if hasattr(obj, 'tolist'):
            # this will work for array.array and numpy.ndarray
            return obj.tolist()
        return JSONEncoder.default(self, obj)



##################################################################
# Copyright 2018 Open Source Geospatial Foundation and others    #
# licensed under MIT, Please consult LICENSE.txt for details     #
##################################################################

"""Validation modes
"""


class MODE():
    """Validation mode enumeration
    """
    NONE = 0
    SIMPLE = 1
    STRICT = 2
    VERYSTRICT = 3



##################################################################
# Copyright 2018 Open Source Geospatial Foundation and others    #
# licensed under MIT, Please consult LICENSE.txt for details     #
##################################################################

from typing import Optional, Dict

Translations = Optional[Dict[str, Dict[str, str]]]



##################################################################
# Copyright 2018 Open Source Geospatial Foundation and others    #
# licensed under MIT, Please consult LICENSE.txt for details     #
##################################################################

import os
import tempfile
import pywps.configuration as config
from pywps import Process, WPSRequest
from pywps.response.execute import ExecuteResponse

import json

import logging
LOGGER = logging.getLogger("PYWPS")


class Job(object):
    """
    :class:`Job` represents a processing job.
    """
    def __init__(self, process, wps_request, wps_response):
        self.process = process
        self.method = '_run_process'
        self.wps_request = wps_request
        self.wps_response = wps_response

    @property
    def name(self):
        return self.process.identifier

    @property
    def workdir(self):
        return self.process.workdir

    @property
    def uuid(self):
        return self.process.uuid

    @property
    def json(self):
        """Return JSON encoded representation of the request
        """
        obj = {
            'process': self.process.json,
            'wps_request': self.wps_request.json,
        }

        return json.dumps(obj, allow_nan=False)

    @classmethod
    def from_json(cls, value):
        """init this request from json back again

        :param value: the json (not string) representation
        """
        process = Process.from_json(value['process'])
        wps_request = WPSRequest()
        wps_request.json = json.loads(value['wps_request'])
        wps_response = ExecuteResponse(
            wps_request=wps_request,
            uuid=process.uuid,
            process=process)
        wps_response.store_status_file = True
        new_job = Job(
            process=Process.from_json(value['process']),
            wps_request=wps_request,
            wps_response=wps_response)
        return new_job

    def dump(self):
        LOGGER.debug('dump job ...')
        filename = tempfile.mkstemp(prefix='job_', suffix='.dump', dir=self.workdir)[1]
        with open(filename, 'w') as fp:
            fp.write(self.json)
            LOGGER.debug("dumped job status to {}".format(filename))
            return filename
        return None

    @classmethod
    def load(cls, filename):
        LOGGER.debug('load job ...')
        with open(filename, 'r') as fp:
            job = Job.from_json(json.load(fp))
            return job
        return None

    def run(self):
        getattr(self.process, self.method)(self.wps_request, self.wps_response)


class JobLauncher(object):
    """
    :class:`JobLauncher` is a command line tool to launch a job from a file
    with a dumped job state.

    Example call: ``joblauncher -c /etc/pywps.cfg job-1001.dump``
    """
    def create_parser(self):
        import argparse
        parser = argparse.ArgumentParser(prog="joblauncher")
        parser.add_argument("-c", "--config", help="Path to pywps configuration.")
        parser.add_argument("filename", help="File with dumped pywps job object.")
        return parser

    def run(self, args):
        if args.config:
            LOGGER.debug("using pywps_cfg={}".format(args.config))
            os.environ['PYWPS_CFG'] = args.config
        self._run_job(args.filename)

    def _run_job(self, filename):
        job = Job.load(filename)
        # init config
        if 'PYWPS_CFG' in os.environ:
            config.load_configuration(os.environ['PYWPS_CFG'])
        # update PATH
        os.environ['PATH'] = "{0}:{1}".format(
            config.get_config_value('processing', 'path'),
            os.environ.get('PATH'))
        # cd into workdir
        os.chdir(job.workdir)
        # init logger ... code copied from app.Service
        if config.get_config_value('logging', 'file') and config.get_config_value('logging', 'level'):
            LOGGER.setLevel(getattr(logging, config.get_config_value('logging', 'level')))
            if not LOGGER.handlers:  # hasHandlers in Python 3.x
                fh = logging.FileHandler(config.get_config_value('logging', 'file'))
                fh.setFormatter(logging.Formatter(config.get_config_value('logging', 'format')))
                LOGGER.addHandler(fh)
        else:  # NullHandler
            if not LOGGER.handlers:
                LOGGER.addHandler(logging.NullHandler())
        job.run()


def launcher():
    """
    Run job launcher command line.
    """
    job_launcher = JobLauncher()
    parser = job_launcher.create_parser()
    args = parser.parse_args()
    job_launcher.run(args)



##################################################################
# Copyright 2018 Open Source Geospatial Foundation and others    #
# licensed under MIT, Please consult LICENSE.txt for details     #
##################################################################

from pywps.validator.mode import MODE


def emptyvalidator(data_input, mode):
    """Empty validator will return always false for security reason
    """

    if mode <= MODE.NONE:
        return True
    else:
        return False



##################################################################
# Copyright 2018 Open Source Geospatial Foundation and others    #
# licensed under MIT, Please consult LICENSE.txt for details     #
##################################################################

import warnings

try:
    import osgeo
    from osgeo import gdal, ogr
except ImportError:
    warnings.warn('Complex validation requires GDAL/OGR support.')
    ogr = None

try:
    import netCDF4
except ImportError:
    warnings.warn('Complex validation requires netCDF4 support.')



##################################################################
# Copyright 2018 Open Source Geospatial Foundation and others    #
# licensed under MIT, Please consult LICENSE.txt for details     #
##################################################################

import logging

import lxml
from pywps import xml_util as etree
from werkzeug.exceptions import MethodNotAllowed
from pywps import get_ElementMakerForVersion
import base64
import datetime
from pywps.app.basic import get_xpath_ns, parse_http_url
from pywps.inout.inputs import input_from_json
from pywps.exceptions import NoApplicableCode, OperationNotSupported, MissingParameterValue, VersionNegotiationFailed, \
    InvalidParameterValue, FileSizeExceeded
from pywps import configuration
from pywps.configuration import wps_strict
from pywps import get_version_from_ns

import json
from urllib.parse import unquote

LOGGER = logging.getLogger("PYWPS")
default_version = '1.0.0'


class WPSRequest(object):

    def __init__(self, http_request=None, preprocessors=None):
        self.http_request = http_request

        self.operation = None
        self.version = None
        self.api = None
        self.default_mimetype = None
        self.language = None
        self.identifier = None
        self.identifiers = None
        self.store_execute = None
        self.status = None
        self.lineage = None
        self.inputs = {}
        self.output_ids = None
        self.outputs = {}
        self.raw = None
        self.WPS = None
        self.OWS = None
        self.xpath_ns = None
        self.preprocessors = preprocessors or dict()
        self.preprocess_request = None
        self.preprocess_response = None

        if http_request:
            d = parse_http_url(http_request)
            self.operation = d.get('operation')
            self.identifier = d.get('identifier')
            self.output_ids = d.get('output_ids')
            self.api = d.get('api')
            self.default_mimetype = d.get('default_mimetype')
            request_parser = self._get_request_parser_method(http_request.method)
            request_parser()

    def _get_request_parser_method(self, method):

        if method == 'GET':
            return self._get_request
        elif method == 'POST':
            return self._post_request
        else:
            raise MethodNotAllowed()

    def _get_request(self):
        """HTTP GET request parser
        """

        # service shall be WPS
        service = _get_get_param(self.http_request, 'service', None if wps_strict else 'wps')
        if service:
            if str(service).lower() != 'wps':
                raise InvalidParameterValue(
                    'parameter SERVICE [{}] not supported'.format(service), 'service')
        else:
            raise MissingParameterValue('service', 'service')

        self.operation = _get_get_param(self.http_request, 'request', self.operation)

        language = _get_get_param(self.http_request, 'language')
        self.check_and_set_language(language)

        request_parser = self._get_request_parser(self.operation)
        request_parser(self.http_request)

    def _post_request(self):
        """HTTP GET request parser
        """
        # check if input file size was not exceeded
        maxsize = configuration.get_config_value('server', 'maxrequestsize')
        maxsize = configuration.get_size_mb(maxsize) * 1024 * 1024
        if self.http_request.content_length > maxsize:
            raise FileSizeExceeded('File size for input exceeded.'
                                   ' Maximum request size allowed: {} megabytes'.format(maxsize / 1024 / 1024))

        content_type = self.http_request.content_type or []  # or self.http_request.mimetype
        json_input = 'json' in content_type
        if not json_input:
            try:
                doc = etree.fromstring(self.http_request.get_data())
            except Exception as e:
                raise NoApplicableCode(e.msg)
            operation = doc.tag
            version = get_version_from_ns(doc.nsmap[doc.prefix])
            self.set_version(version)

            language = doc.attrib.get('language')
            self.check_and_set_language(language)

            request_parser = self._post_request_parser(operation)
            request_parser(doc)
        else:
            try:
                jdoc = json.loads(self.http_request.get_data())
            except Exception as e:
                raise NoApplicableCode(e.msg)
            if self.identifier is not None:
                jdoc = {'inputs': jdoc}
            else:
                self.identifier = jdoc.get('identifier', None)

            self.operation = jdoc.get('operation', self.operation)

            preprocessor_tuple = self.preprocessors.get(self.identifier, None)
            if preprocessor_tuple:
                self.identifier = preprocessor_tuple[0]
                self.preprocess_request = preprocessor_tuple[1]
                self.preprocess_response = preprocessor_tuple[2]

            jdoc['operation'] = self.operation
            jdoc['identifier'] = self.identifier
            jdoc['api'] = self.api
            jdoc['default_mimetype'] = self.default_mimetype

            if self.preprocess_request is not None:
                jdoc = self.preprocess_request(jdoc, http_request=self.http_request)
            self.json = jdoc

            version = jdoc.get('version')
            self.set_version(version)

            language = jdoc.get('language')
            self.check_and_set_language(language)

            request_parser = self._post_json_request_parser()
            request_parser(jdoc)

    def _get_request_parser(self, operation):
        """Factory function returing propper parsing function
        """

        wpsrequest = self

        def parse_get_getcapabilities(http_request):
            """Parse GET GetCapabilities request
            """

            acceptedversions = _get_get_param(http_request, 'acceptversions')
            wpsrequest.check_accepted_versions(acceptedversions)
            wpsrequest.default_mimetype = _get_get_param(http_request, 'f', wpsrequest.default_mimetype)

        def parse_get_describeprocess(http_request):
            """Parse GET DescribeProcess request
            """
            version = _get_get_param(http_request, 'version')
            wpsrequest.check_and_set_version(version)

            wpsrequest.identifiers = _get_get_param(
                http_request, 'identifier', wpsrequest.identifiers, aslist=True)
            if wpsrequest.identifiers is None and self.identifier is not None:
                wpsrequest.identifiers = [wpsrequest.identifier]
            wpsrequest.default_mimetype = _get_get_param(http_request, 'f', wpsrequest.default_mimetype)

        def parse_get_execute(http_request):
            """Parse GET Execute request
            """
            version = _get_get_param(http_request, 'version')
            wpsrequest.check_and_set_version(version)

            wpsrequest.identifier = _get_get_param(http_request, 'identifier', wpsrequest.identifier)
            wpsrequest.store_execute = _get_get_param(
                http_request, 'storeExecuteResponse', 'false')
            wpsrequest.status = _get_get_param(http_request, 'status', 'false')
            wpsrequest.lineage = _get_get_param(
                http_request, 'lineage', 'false')
            wpsrequest.inputs = get_data_from_kvp(
                _get_get_param(http_request, 'DataInputs'), 'DataInputs')
            if self.inputs is None:
                self.inputs = {}

            # take responseDocument preferably
            raw, output_ids = False, _get_get_param(http_request, 'ResponseDocument')
            if output_ids is None:
                raw, output_ids = True, _get_get_param(http_request, 'RawDataOutput')
            if output_ids is not None:
                wpsrequest.raw, wpsrequest.output_ids = raw, output_ids
            elif wpsrequest.raw is None:
                wpsrequest.raw = wpsrequest.output_ids is not None

            wpsrequest.default_mimetype = _get_get_param(http_request, 'f', wpsrequest.default_mimetype)
            wpsrequest.outputs = get_data_from_kvp(wpsrequest.output_ids) or {}
            if wpsrequest.raw:
                # executeResponse XML will not be stored and no updating of
                # status
                wpsrequest.store_execute = 'false'
                wpsrequest.status = 'false'

        if operation:
            self.operation = operation.lower()
        else:
            if wps_strict:
                raise MissingParameterValue('Missing request value', 'request')
            self.operation = 'execute'

        if self.operation == 'getcapabilities':
            return parse_get_getcapabilities
        elif self.operation == 'describeprocess':
            return parse_get_describeprocess
        elif self.operation == 'execute':
            return parse_get_execute
        else:
            raise OperationNotSupported(
                'Unknown request {}'.format(self.operation), operation)

    def _post_request_parser(self, tagname):
        """Factory function returning a proper parsing function
        according to tagname and sets self.operation to the correct operation
        """

        wpsrequest = self

        def parse_post_getcapabilities(doc):
            """Parse POST GetCapabilities request
            """
            acceptedversions = self.xpath_ns(
                doc, '/wps:GetCapabilities/ows:AcceptVersions/ows:Version')
            acceptedversions = ','.join(
                [v.text for v in acceptedversions])
            wpsrequest.check_accepted_versions(acceptedversions)

        def parse_post_describeprocess(doc):
            """Parse POST DescribeProcess request
            """

            version = doc.attrib.get('version')
            wpsrequest.check_and_set_version(version)

            wpsrequest.operation = 'describeprocess'
            wpsrequest.identifiers = [identifier_el.text for identifier_el in
                                      self.xpath_ns(doc, './ows:Identifier')]

        def parse_post_execute(doc):
            """Parse POST Execute request
            """
            version = doc.attrib.get('version')
            wpsrequest.check_and_set_version(version)

            wpsrequest.operation = 'execute'

            identifier = self.xpath_ns(doc, './ows:Identifier')

            if not identifier:
                raise MissingParameterValue(
                    'Process identifier not set', 'Identifier')

            wpsrequest.identifier = identifier[0].text
            wpsrequest.lineage = 'false'
            wpsrequest.store_execute = 'false'
            wpsrequest.status = 'false'
            wpsrequest.inputs = get_inputs_from_xml(doc)
            wpsrequest.outputs = get_output_from_xml(doc)
            wpsrequest.raw = False
            if self.xpath_ns(doc, '/wps:Execute/wps:ResponseForm/wps:RawDataOutput'):
                wpsrequest.raw = True
                # executeResponse XML will not be stored
                wpsrequest.store_execute = 'false'

            # check if response document tag has been set then retrieve
            response_document = self.xpath_ns(
                doc, './wps:ResponseForm/wps:ResponseDocument')
            if len(response_document) > 0:
                wpsrequest.lineage = response_document[
                    0].attrib.get('lineage', 'false')
                wpsrequest.store_execute = response_document[
                    0].attrib.get('storeExecuteResponse', 'false')
                wpsrequest.status = response_document[
                    0].attrib.get('status', 'false')

        if tagname == self.WPS.GetCapabilities().tag:
            self.operation = 'getcapabilities'
            return parse_post_getcapabilities
        elif tagname == self.WPS.DescribeProcess().tag:
            self.operation = 'describeprocess'
            return parse_post_describeprocess
        elif tagname == self.WPS.Execute().tag:
            self.operation = 'execute'
            return parse_post_execute
        else:
            raise InvalidParameterValue(
                'Unknown request {}'.format(tagname), 'request')

    def _post_json_request_parser(self):
        """
        Factory function returning a proper parsing function
        according to self.operation.
        self.operation is modified to be lowercase
            or the default 'execute' operation if self.operation is None
        """

        wpsrequest = self

        def parse_json_post_getcapabilities(jdoc):
            """Parse POST GetCapabilities request
            """
            acceptedversions = jdoc.get('acceptedversions')
            wpsrequest.check_accepted_versions(acceptedversions)

        def parse_json_post_describeprocess(jdoc):
            """Parse POST DescribeProcess request
            """

            version = jdoc.get('version')
            wpsrequest.check_and_set_version(version)
            wpsrequest.identifiers = [identifier_el.text for identifier_el in
                                      self.xpath_ns(jdoc, './ows:Identifier')]

        def parse_json_post_execute(jdoc):
            """Parse POST Execute request
            """
            version = jdoc.get('version')
            wpsrequest.check_and_set_version(version)

            wpsrequest.identifier = jdoc.get('identifier')
            if wpsrequest.identifier is None:
                raise MissingParameterValue(
                    'Process identifier not set', 'Identifier')

            wpsrequest.lineage = 'false'
            wpsrequest.store_execute = 'false'
            wpsrequest.status = 'false'
            wpsrequest.inputs = get_inputs_from_json(jdoc)

            if wpsrequest.output_ids is None:
                wpsrequest.output_ids = jdoc.get('outputs', {})
                wpsrequest.raw = jdoc.get('raw', False)
            wpsrequest.raw, wpsrequest.outputs = get_output_from_dict(wpsrequest.output_ids, wpsrequest.raw)

            if wpsrequest.raw:
                # executeResponse XML will not be stored
                wpsrequest.store_execute = 'false'

            # todo: parse response_document like in the xml version?

        self.operation = 'execute' if self.operation is None else self.operation.lower()
        if self.operation == 'getcapabilities':
            return parse_json_post_getcapabilities
        elif self.operation == 'describeprocess':
            return parse_json_post_describeprocess
        elif self.operation == 'execute':
            return parse_json_post_execute
        else:
            raise InvalidParameterValue(
                'Unknown request {}'.format(self.operation), 'request')

    def set_version(self, version):
        self.version = version
        self.xpath_ns = get_xpath_ns(version)
        self.WPS, self.OWS = get_ElementMakerForVersion(self.version)

    def check_accepted_versions(self, acceptedversions):
        """
        :param acceptedversions: string
        """

        version = None

        if acceptedversions:
            acceptedversions_array = acceptedversions.split(',')
            for aversion in acceptedversions_array:
                if _check_version(aversion):
                    version = aversion
        else:
            version = '1.0.0'

        if version:
            self.check_and_set_version(version)
        else:
            raise VersionNegotiationFailed(
                'The requested version "{}" is not supported by this server'.format(acceptedversions), 'version')

    def check_and_set_version(self, version, allow_default=True):
        """set this.version
        """

        if not version:
            if allow_default:
                version = default_version
            else:
                raise MissingParameterValue('Missing version', 'version')
        if not _check_version(version):
            raise VersionNegotiationFailed(
                'The requested version "{}" is not supported by this server'.format(version), 'version')
        else:
            self.set_version(version)

    def check_and_set_language(self, language):
        """set this.language
        """
        supported_languages = configuration.get_config_value('server', 'language').split(',')
        supported_languages = [lang.strip() for lang in supported_languages]

        if not language:
            # default to the first supported language
            language = supported_languages[0]

        if language not in supported_languages:
            raise InvalidParameterValue(
                'The requested language "{}" is not supported by this server'.format(language),
                'language',
            )

        self.language = language

    @property
    def json(self):
        """Return JSON encoded representation of the request
        """
        class ExtendedJSONEncoder(json.JSONEncoder):
            def default(self, obj):
                if isinstance(obj, datetime.date) or isinstance(obj, datetime.time):
                    encoded_object = obj.isoformat()
                else:
                    encoded_object = json.JSONEncoder.default(self, obj)
                return encoded_object

        obj = {
            'operation': self.operation,
            'version': self.version,
            'api': self.api,
            'default_mimetype': self.default_mimetype,
            'language': self.language,
            'identifier': self.identifier,
            'identifiers': self.identifiers,
            'store_execute': self.store_execute,
            'status': self.status,
            'lineage': self.lineage,
            'inputs': dict((i, [inpt.json for inpt in self.inputs[i]]) for i in self.inputs),
            'outputs': self.outputs,
            'raw': self.raw
        }

        return json.dumps(obj, allow_nan=False, cls=ExtendedJSONEncoder)

    @json.setter
    def json(self, value):
        """init this request from json back again

        :param value: the json (not string) representation
        """

        self.operation = value.get('operation')
        self.version = value.get('version')
        self.api = value.get('api')
        self.default_mimetype = value.get('default_mimetype')
        self.language = value.get('language')
        self.identifier = value.get('identifier')
        self.identifiers = value.get('identifiers')
        self.store_execute = value.get('store_execute')
        self.status = value.get('status', False)
        self.lineage = value.get('lineage', False)
        self.outputs = value.get('outputs')
        self.raw = value.get('raw', False)
        self.inputs = {}

        for identifier in value.get('inputs', []):
            inpt_defs = value['inputs'][identifier]
            if not isinstance(inpt_defs, (list, tuple)):
                inpt_defs = [inpt_defs]
            self.inputs[identifier] = []
            for inpt_def in inpt_defs:
                if not isinstance(inpt_def, dict):
                    inpt_def = {"data": inpt_def}
                if 'identifier' not in inpt_def:
                    inpt_def['identifier'] = identifier
                inpt = input_from_json(inpt_def)
                self.inputs[identifier].append(inpt)


def get_inputs_from_xml(doc):
    the_inputs = {}
    version = get_version_from_ns(doc.nsmap[doc.prefix])
    xpath_ns = get_xpath_ns(version)
    for input_el in xpath_ns(doc, '/wps:Execute/wps:DataInputs/wps:Input'):
        [identifier_el] = xpath_ns(input_el, './ows:Identifier')
        identifier = identifier_el.text

        if identifier not in the_inputs:
            the_inputs[identifier] = []

        literal_data = xpath_ns(input_el, './wps:Data/wps:LiteralData')
        if literal_data:
            value_el = literal_data[0]
            inpt = {}
            inpt['identifier'] = identifier_el.text
            inpt['data'] = str(value_el.text)
            inpt['uom'] = value_el.attrib.get('uom', '')
            inpt['datatype'] = value_el.attrib.get('datatype', '')
            the_inputs[identifier].append(inpt)
            continue

        complex_data = xpath_ns(input_el, './wps:Data/wps:ComplexData')
        if complex_data:
            complex_data_el = complex_data[0]
            inpt = {}
            inpt['identifier'] = identifier_el.text
            inpt['mimeType'] = complex_data_el.attrib.get('mimeType', None)
            inpt['encoding'] = complex_data_el.attrib.get('encoding', '').lower()
            inpt['schema'] = complex_data_el.attrib.get('schema', '')
            inpt['method'] = complex_data_el.attrib.get('method', 'GET')
            if len(complex_data_el.getchildren()) > 0:
                value_el = complex_data_el[0]
                inpt['data'] = _get_dataelement_value(value_el)
            else:
                inpt['data'] = _get_rawvalue_value(
                    complex_data_el.text, inpt['encoding'])
            the_inputs[identifier].append(inpt)
            continue

        reference_data = xpath_ns(input_el, './wps:Reference')
        if reference_data:
            reference_data_el = reference_data[0]
            inpt = {}
            inpt['identifier'] = identifier_el.text
            inpt[identifier_el.text] = reference_data_el.text
            inpt['href'] = reference_data_el.attrib.get(
                '{http://www.w3.org/1999/xlink}href', '')
            inpt['mimeType'] = reference_data_el.attrib.get('mimeType', None)
            inpt['method'] = reference_data_el.attrib.get('method', 'GET')
            header_element = xpath_ns(reference_data_el, './wps:Header')
            if header_element:
                inpt['header'] = _get_reference_header(header_element)
            body_element = xpath_ns(reference_data_el, './wps:Body')
            if body_element:
                inpt['body'] = _get_reference_body(body_element[0])
            bodyreference_element = xpath_ns(reference_data_el,
                                             './wps:BodyReference')
            if bodyreference_element:
                inpt['bodyreference'] = _get_reference_bodyreference(
                    bodyreference_element[0])
            the_inputs[identifier].append(inpt)
            continue

        # Using OWSlib BoundingBox
        from owslib.ows import BoundingBox
        bbox_datas = xpath_ns(input_el, './wps:Data/wps:BoundingBoxData')
        if bbox_datas:
            for bbox_data in bbox_datas:
                bbox = BoundingBox(bbox_data)
                LOGGER.debug("parse bbox: minx={}, miny={}, maxx={},maxy={}".format(
                    bbox.minx, bbox.miny, bbox.maxx, bbox.maxy))
                inpt = {}
                inpt['identifier'] = identifier_el.text
                inpt['data'] = [bbox.minx, bbox.miny, bbox.maxx, bbox.maxy]
                inpt['crs'] = bbox.crs
                inpt['dimensions'] = bbox.dimensions
                the_inputs[identifier].append(inpt)
    return the_inputs


def get_output_from_xml(doc):
    the_output = {}

    version = get_version_from_ns(doc.nsmap[doc.prefix])
    xpath_ns = get_xpath_ns(version)

    if xpath_ns(doc, '/wps:Execute/wps:ResponseForm/wps:ResponseDocument'):
        for output_el in xpath_ns(doc, '/wps:Execute/wps:ResponseForm/wps:ResponseDocument/wps:Output'):
            [identifier_el] = xpath_ns(output_el, './ows:Identifier')
            outpt = {}
            outpt[identifier_el.text] = ''
            outpt['mimetype'] = output_el.attrib.get('mimeType', None)
            outpt['encoding'] = output_el.attrib.get('encoding', '')
            outpt['schema'] = output_el.attrib.get('schema', '')
            outpt['uom'] = output_el.attrib.get('uom', '')
            outpt['asReference'] = output_el.attrib.get('asReference', 'false')
            the_output[identifier_el.text] = outpt

    elif xpath_ns(doc, '/wps:Execute/wps:ResponseForm/wps:RawDataOutput'):
        for output_el in xpath_ns(doc, '/wps:Execute/wps:ResponseForm/wps:RawDataOutput'):
            [identifier_el] = xpath_ns(output_el, './ows:Identifier')
            outpt = {}
            outpt[identifier_el.text] = ''
            outpt['mimetype'] = output_el.attrib.get('mimeType', None)
            outpt['encoding'] = output_el.attrib.get('encoding', '')
            outpt['schema'] = output_el.attrib.get('schema', '')
            outpt['uom'] = output_el.attrib.get('uom', '')
            the_output[identifier_el.text] = outpt

    return the_output


def get_inputs_from_json(jdoc):
    the_inputs = {}
    inputs_dict = jdoc.get('inputs', {})
    for identifier, inpt_defs in inputs_dict.items():
        if not isinstance(inpt_defs, (list, tuple)):
            inpt_defs = [inpt_defs]
        the_inputs[identifier] = []
        for inpt_def in inpt_defs:
            if not isinstance(inpt_def, dict):
                inpt_def = {"data": inpt_def}
            data_type = inpt_def.get('type', 'literal')
            inpt = {'identifier': identifier}
            if data_type == 'literal':
                inpt['data'] = inpt_def.get('data')
                inpt['uom'] = inpt_def.get('uom', '')
                inpt['datatype'] = inpt_def.get('datatype', '')
                the_inputs[identifier].append(inpt)
            elif data_type == 'complex':
                inpt['mimeType'] = inpt_def.get('mimeType', None)
                inpt['encoding'] = inpt_def.get('encoding', '').lower()
                inpt['schema'] = inpt_def.get('schema', '')
                inpt['method'] = inpt_def.get('method', 'GET')
                inpt['data'] = _get_rawvalue_value(inpt_def.get('data', ''), inpt['encoding'])
                the_inputs[identifier].append(inpt)
            elif data_type == 'reference':
                inpt[identifier] = inpt_def
                inpt['href'] = inpt_def.get('href', '')
                inpt['mimeType'] = inpt_def.get('mimeType', None)
                inpt['method'] = inpt_def.get('method', 'GET')
                inpt['header'] = inpt_def.get('header', '')
                inpt['body'] = inpt_def.get('body', '')
                inpt['bodyreference'] = inpt_def.get('bodyreference', '')
                the_inputs[identifier].append(inpt)
            elif data_type == 'bbox':
                inpt['data'] = inpt_def['bbox']
                inpt['crs'] = inpt_def.get('crs', 'urn:ogc:def:crs:EPSG::4326')
                inpt['dimensions'] = inpt_def.get('dimensions', 2)
                the_inputs[identifier].append(inpt)
    return the_inputs


def get_output_from_dict(output_ids, raw):
    the_output = {}
    if isinstance(output_ids, dict):
        pass
    elif isinstance(output_ids, (tuple, list)):
        output_ids = {x: {} for x in output_ids}
    else:
        output_ids = {output_ids: {}}
        raw = True  # single non-dict output means raw output
    for identifier, output_el in output_ids.items():
        if isinstance(output_el, list):
            output_el = output_el[0]
        outpt = {}
        outpt[identifier] = ''
        outpt['mimetype'] = output_el.get('mimeType', None)
        outpt['encoding'] = output_el.get('encoding', '')
        outpt['schema'] = output_el.get('schema', '')
        outpt['uom'] = output_el.get('uom', '')
        if not raw:
            outpt['asReference'] = output_el.get('asReference', 'false')
        the_output[identifier] = outpt

    return raw, the_output


def get_data_from_kvp(data, part=None):
    """Get execute DataInputs and ResponseDocument from URL (key-value-pairs) encoding
    :param data: key:value pair list of the datainputs and responseDocument parameter
    :param part: DataInputs or similar part of input url
    """

    the_data = {}

    if data is None:
        return None

    for d in data.split(";"):
        try:
            io = {}
            fields = d.split('@')

            # First field is identifier and its value
            (identifier, val) = fields[0].split("=")
            io['identifier'] = identifier
            io['data'] = unquote(val)

            # Get the attributes of the data
            for attr in fields[1:]:
                (attribute, attr_val) = attr.split('=', 1)
                if attribute == 'xlink:href':
                    io['href'] = unquote(attr_val)
                else:
                    io[attribute] = unquote(attr_val)

            # Add the input/output with all its attributes and values to the
            # dictionary
            if part == 'DataInputs':
                if identifier not in the_data:
                    the_data[identifier] = []
                the_data[identifier].append(io)
            else:
                the_data[identifier] = io
        except Exception as e:
            LOGGER.warning(e)
            the_data[d] = {'identifier': d, 'data': ''}

    return the_data


def _check_version(version):
    """ check given version
    """
    if version not in ['1.0.0', '2.0.0']:
        return False
    else:
        return True


def _get_get_param(http_request, key, default=None, aslist=False):
    """Returns value from the key:value pair, of the HTTP GET request, for
    example 'service' or 'request'

    :param http_request: http_request object
    :param key: key value you need to dig out of the HTTP GET request
    """

    key = key.lower()
    value = default
    # http_request.args.keys will make + sign disappear in GET url if not
    # urlencoded
    for k in http_request.args.keys():
        if k.lower() == key:
            value = http_request.args.get(k)
            if aslist:
                value = value.split(",")

    return value


def _get_dataelement_value(value_el):
    """Return real value of XML Element (e.g. convert Element.FeatureCollection
    to String
    """

    if isinstance(value_el, lxml.etree._Element):
        return etree.tostring(value_el, encoding=str)
    else:
        return value_el


def _get_rawvalue_value(data, encoding=None):
    """Return real value of CDATA section"""

    try:
        LOGGER.debug("encoding={}".format(encoding))
        if encoding is None or encoding == "":
            return data
        elif encoding == "utf-8":
            return data
        elif encoding == 'base64':
            return base64.b64decode(data)
        return base64.b64decode(data)
    except Exception:
        LOGGER.warning("failed to decode base64")
        return data


def _get_reference_header(header_element):
    """Parses ReferenceInput Header element
    """
    header = {}
    header['key'] = header_element.attrib('key')
    header['value'] = header_element.attrib('value')
    return header


def _get_reference_body(body_element):
    """Parses ReferenceInput Body element
    """

    body = None
    if len(body_element.getchildren()) > 0:
        value_el = body_element[0]
        body = _get_dataelement_value(value_el)
    else:
        body = _get_rawvalue_value(body_element.text)

    return body


def _get_reference_bodyreference(referencebody_element):
    """Parse ReferenceInput BodyReference element
    """
    return referencebody_element.attrib.get(
        '{http://www.w3.org/1999/xlink}href', '')



##################################################################
# Copyright 2018 Open Source Geospatial Foundation and others    #
# licensed under MIT, Please consult LICENSE.txt for details     #
##################################################################

"""
Reads the PyWPS configuration file
"""

import logging
import sys
import os
import tempfile
import pywps

import configparser

__author__ = "Calin Ciociu"

from pywps.util import file_uri

RAW_OPTIONS = [('logging', 'format'), ]

CONFIG = None
LOGGER = logging.getLogger("PYWPS")

wps_strict = True


def get_config_value(section, option, default_value=''):
    """Get desired value from  configuration files

    :param section: section in configuration files
    :type section: string
    :param option: option in the section
    :type option: string
    :returns: value found in the configuration file
    """

    if not CONFIG:
        load_configuration()

    value = default_value

    if CONFIG.has_section(section):
        if CONFIG.has_option(section, option):
            raw = (section, option) in RAW_OPTIONS
            value = CONFIG.get(section, option, raw=raw)

            # Convert Boolean string to real Boolean values
            if value.lower() == "false":
                value = False
            elif value.lower() == "true":
                value = True

    return value


def load_configuration(cfgfiles=None):
    """Load PyWPS configuration from configuration files.
    The later configuration file in the array overwrites configuration
    from the first.

    :param cfgfiles: list of configuration files
    """

    global CONFIG

    LOGGER.info('loading configuration')
    CONFIG = configparser.ConfigParser(os.environ)

    LOGGER.debug('setting default values')
    CONFIG.add_section('server')
    CONFIG.set('server', 'encoding', 'utf-8')
    CONFIG.set('server', 'language', 'en-US')
    CONFIG.set('server', 'url', 'http://localhost/wps')
    CONFIG.set('server', 'maxprocesses', '30')
    CONFIG.set('server', 'maxsingleinputsize', '1mb')
    CONFIG.set('server', 'maxrequestsize', '3mb')
    CONFIG.set('server', 'temp_path', tempfile.gettempdir())
    CONFIG.set('server', 'processes_path', '')
    outputpath = tempfile.gettempdir()
    CONFIG.set('server', 'outputurl', file_uri(outputpath))
    CONFIG.set('server', 'outputpath', outputpath)
    # list of allowed input paths (file url input) seperated by ':'
    CONFIG.set('server', 'allowedinputpaths', '')
    CONFIG.set('server', 'workdir', tempfile.gettempdir())
    CONFIG.set('server', 'parallelprocesses', '2')
    # If this flag is enabled it will set the HOME environment
    # for each process to its current workdir (a temp folder).
    CONFIG.set('server', 'sethomedir', 'false')
    # If this flag is enabled PyWPS will remove the process temporary workdir
    # after process has finished.
    CONFIG.set('server', 'cleantempdir', 'true')
    CONFIG.set('server', 'storagetype', 'file')
    # File storage outputs can be copied, moved or linked
    # from the workdir to the output folder.
    # Allowed functions: "copy", "move", "link" (default "copy")
    CONFIG.set('server', 'storage_copy_function', 'copy')

    # handles the default mimetype for requests.
    # available options: "text/xml", "application/json"
    CONFIG.set("server", "default_mimetype", "text/xml")

    # default json indentation for responses.
    CONFIG.set("server", "json_indent", "2")

    CONFIG.add_section('processing')
    CONFIG.set('processing', 'mode', 'default')
    CONFIG.set('processing', 'path', os.path.dirname(os.path.realpath(sys.argv[0])))
    # https://github.com/natefoo/slurm-drmaa
    CONFIG.set('processing', 'drmaa_native_specification', '')

    CONFIG.add_section('logging')
    CONFIG.set('logging', 'file', '')
    CONFIG.set('logging', 'level', 'DEBUG')
    CONFIG.set('logging', 'database', 'sqlite:///:memory:')
    CONFIG.set('logging', 'prefix', 'pywps_')
    CONFIG.set('logging', 'format', '%(asctime)s] [%(levelname)s] file=%(pathname)s line=%(lineno)s module=%(module)s function=%(funcName)s %(message)s')  # noqa

    CONFIG.add_section('metadata:main')
    CONFIG.set('metadata:main', 'identification_title', 'PyWPS Processing Service')
    CONFIG.set('metadata:main', 'identification_abstract', 'PyWPS is an implementation of the Web Processing Service standard from the Open Geospatial Consortium. PyWPS is written in Python.')  # noqa
    CONFIG.set('metadata:main', 'identification_keywords', 'PyWPS,WPS,OGC,processing')
    CONFIG.set('metadata:main', 'identification_keywords_type', 'theme')
    CONFIG.set('metadata:main', 'identification_fees', 'NONE')
    CONFIG.set('metadata:main', 'identification_accessconstraints', 'NONE')
    CONFIG.set('metadata:main', 'provider_name', 'Organization Name')
    CONFIG.set('metadata:main', 'provider_url', 'https://pywps.org/')
    CONFIG.set('metadata:main', 'contact_name', 'Lastname, Firstname')
    CONFIG.set('metadata:main', 'contact_position', 'Position Title')
    CONFIG.set('metadata:main', 'contact_address', 'Mailing Address')
    CONFIG.set('metadata:main', 'contact_city', 'City')
    CONFIG.set('metadata:main', 'contact_stateorprovince', 'Administrative Area')
    CONFIG.set('metadata:main', 'contact_postalcode', 'Zip or Postal Code')
    CONFIG.set('metadata:main', 'contact_country', 'Country')
    CONFIG.set('metadata:main', 'contact_phone', '+xx-xxx-xxx-xxxx')
    CONFIG.set('metadata:main', 'contact_fax', '+xx-xxx-xxx-xxxx')
    CONFIG.set('metadata:main', 'contact_email', 'Email Address')
    CONFIG.set('metadata:main', 'contact_url', 'Contact URL')
    CONFIG.set('metadata:main', 'contact_hours', 'Hours of Service')
    CONFIG.set('metadata:main', 'contact_instructions', 'During hours of service.  Off on weekends.')
    CONFIG.set('metadata:main', 'contact_role', 'pointOfContact')

    CONFIG.add_section('grass')
    CONFIG.set('grass', 'gisbase', '')

    CONFIG.add_section('s3')
    CONFIG.set('s3', 'bucket', '')
    CONFIG.set('s3', 'prefix', '')
    CONFIG.set('s3', 'public', 'false')
    CONFIG.set('s3', 'encrypt', 'false')
    CONFIG.set('s3', 'region', '')

    if not cfgfiles:
        cfgfiles = _get_default_config_files_location()

    if isinstance(cfgfiles, str):
        cfgfiles = [cfgfiles]

    if 'PYWPS_CFG' in os.environ:
        cfgfiles.append(os.environ['PYWPS_CFG'])

    loaded_files = CONFIG.read(cfgfiles, encoding='utf-8')
    if loaded_files:
        LOGGER.info('Configuration file(s) {} loaded'.format(loaded_files))
    else:
        LOGGER.info('No configuration files loaded. Using default values')

    _check_config()


def _check_config():
    """Check some configuration values
    """
    global CONFIG

    def checkdir(confid):

        confvalue = get_config_value('server', confid)

        if not os.path.isdir(confvalue):
            LOGGER.warning('server->{} configuration value {} is not directory'.format(confid, confvalue))

        if not os.path.isabs(confvalue):
            LOGGER.warning('server->{} configuration value {} is not absolute path, making it absolute to {}'.format(
                           confid, confvalue, os.path.abspath(confvalue)))
            CONFIG.set('server', confid, os.path.abspath(confvalue))

    [checkdir(n) for n in ['workdir', 'outputpath']]


def _get_default_config_files_location():
    """Get the locations of the standard configuration files. These are
    Unix/Linux:
        1. `/etc/pywps.cfg`
        2. `$HOME/.pywps.cfg`
    Windows:
        1. `pywps\\etc\\default.cfg`

    Both:
        1. `$PYWPS_CFG environment variable`
    :returns: configuration files
    :rtype: list of strings
    """

    is_win32 = sys.platform == 'win32'
    if is_win32:
        LOGGER.debug('Windows based environment')
    else:
        LOGGER.debug('UNIX based environment')

    if os.getenv("PYWPS_CFG"):
        LOGGER.debug('using PYWPS_CFG environment variable')
        # Windows or Unix
        if is_win32:
            PYWPS_INSTALL_DIR = os.path.abspath(os.path.join(os.getcwd(), os.path.dirname(sys.argv[0])))
            cfgfiles = (os.getenv("PYWPS_CFG"))
        else:
            cfgfiles = (os.getenv("PYWPS_CFG"))

    else:
        LOGGER.debug('trying to estimate the default location')
        # Windows or Unix
        if is_win32:
            PYWPS_INSTALL_DIR = os.path.abspath(os.path.join(os.getcwd(), os.path.dirname(sys.argv[0])))
            cfgfiles = (os.path.join(PYWPS_INSTALL_DIR, "pywps", "etc", "pywps.cfg"))
        else:
            homePath = os.getenv("HOME")
            if homePath:
                cfgfiles = (os.path.join(pywps.__path__[0], "etc", "pywps.cfg"), "/etc/pywps.cfg",
                            os.path.join(os.getenv("HOME"), ".pywps.cfg"))
            else:
                cfgfiles = (os.path.join(pywps.__path__[0], "etc",
                            "pywps.cfg"), "/etc/pywps.cfg")

    return cfgfiles


def get_size_mb(mbsize):
    """Get real size of given object in Mb.

    """

    size = mbsize.lower()

    import re

    units = re.compile("[gmkb].*")
    newsize = float(re.sub(units, '', size))

    if size.find("g") > -1:
        newsize *= 1024
    elif size.find("m") > -1:
        newsize *= 1
    elif size.find("k") > -1:
        newsize /= 1024
    else:
        newsize *= 1
    LOGGER.debug('Calculated real size of {} is {}'.format(mbsize, newsize))
    return newsize



##################################################################
# Copyright 2018 Open Source Geospatial Foundation and others    #
# licensed under MIT, Please consult LICENSE.txt for details     #
##################################################################

from collections import namedtuple

_ALLOWEDVALUETYPE = namedtuple('ALLOWEDVALUETYPE', 'VALUE, RANGE')
_RANGELCLOSURETYPE = namedtuple('RANGECLOSURETYPE', 'OPEN, CLOSED,'
                                'OPENCLOSED, CLOSEDOPEN')

ALLOWEDVALUETYPE = _ALLOWEDVALUETYPE('value', 'range')
RANGECLOSURETYPE = _RANGELCLOSURETYPE(
    'open',
    'closed',
    'open-closed',
    'closed-open'
)



##################################################################
# Copyright 2018 Open Source Geospatial Foundation and others    #
# licensed under MIT, Please consult LICENSE.txt for details     #
##################################################################

import re
from pywps import xml_util as etree

from pywps.app.Common import Metadata
from pywps.exceptions import InvalidParameterValue
from pywps.inout.formats import Format
from pywps.inout import basic
from copy import deepcopy
from pywps.validator.mode import MODE
from pywps.inout.literaltypes import AnyValue, NoValue, ValuesReference, AllowedValue

CDATA_PATTERN = re.compile(r'<!\[CDATA\[(.*?)\]\]>')


class BoundingBoxInput(basic.BBoxInput):

    """
    :param string identifier: The name of this input.
    :param string title: Human readable title
    :param string abstract: Longer text description
    :param crss: List of supported coordinate reference
                 system (e.g. ['EPSG:4326'])
    :param list keywords: Keywords that characterize this input.
    :param int dimensions: 2 or 3
    :param str workdir: working directory, to save temporary file objects in.
    :param list metadata: TODO
    :param int min_occurs: how many times this input occurs
    :param int max_occurs: how many times this input occurs
    :param metadata: List of metadata advertised by this process. They
                     should be :class:`pywps.app.Common.Metadata` objects.
    :param dict[str,dict[str,str]] translations: The first key is the RFC 4646 language code,
        and the nested mapping contains translated strings accessible by a string property.
        e.g. {"fr-CA": {"title": "Mon titre", "abstract": "Une description"}}
    """

    def __init__(self, identifier, title, crss=None, abstract='', keywords=[],
                 dimensions=2, workdir=None, metadata=[], min_occurs=1,
                 max_occurs=1,
                 mode=MODE.NONE,
                 default=None, default_type=basic.SOURCE_TYPE.DATA,
                 translations=None):

        basic.BBoxInput.__init__(self, identifier, title=title, crss=crss,
                                 abstract=abstract, keywords=keywords,
                                 dimensions=dimensions, workdir=workdir, metadata=metadata,
                                 min_occurs=min_occurs, max_occurs=max_occurs,
                                 mode=mode, default=default,
                                 default_type=default_type,
                                 translations=translations)

        self.as_reference = False

    @property
    def json(self):
        """Get JSON representation of the input
        """
        return {
            'identifier': self.identifier,
            'title': self.title,
            'abstract': self.abstract,
            'keywords': self.keywords,
            'type': 'bbox',
            'crs': self.crs,
            'crss': self.crss,
            'metadata': [m.json for m in self.metadata],
            'bbox': self.data,
            'll': self.ll,
            'ur': self.ur,
            'dimensions': self.dimensions,
            'workdir': self.workdir,
            'mode': self.valid_mode,
            'min_occurs': self.min_occurs,
            'max_occurs': self.max_occurs,
            'translations': self.translations,
        }

    @classmethod
    def from_json(cls, json_input):
        instance = cls(
            identifier=json_input['identifier'],
            title=json_input.get('title'),
            abstract=json_input.get('abstract'),
            crss=json_input.get('crss'),
            keywords=json_input.get('keywords'),
            metadata=[Metadata.from_json(data) for data in json_input.get('metadata', [])],
            dimensions=json_input.get('dimensions'),
            workdir=json_input.get('workdir'),
            mode=json_input.get('mode'),
            min_occurs=json_input.get('min_occurs'),
            max_occurs=json_input.get('max_occurs'),
            translations=json_input.get('translations'),
        )
        instance.data = json_input['bbox']

        return instance

    def clone(self):
        """Create copy of yourself
        """
        return deepcopy(self)


class ComplexInput(basic.ComplexInput):
    """
    Complex data input

    :param str identifier: The name of this input.
    :param str title: Title of the input
    :param pywps.inout.formats.Format supported_formats: List of supported
                                                          formats
    :param pywps.inout.formats.Format data_format: default data format
    :param str abstract: Input abstract
    :param list keywords: Keywords that characterize this input.
    :param str workdir: working directory, to save temporary file objects in.
    :param list metadata: TODO
    :param int min_occurs: minimum occurrence
    :param int max_occurs: maximum occurrence
    :param pywps.validator.mode.MODE mode: validation mode (none to strict)
    :param dict[str,dict[str,str]] translations: The first key is the RFC 4646 language code,
        and the nested mapping contains translated strings accessible by a string property.
        e.g. {"fr-CA": {"title": "Mon titre", "abstract": "Une description"}}
    """

    def __init__(self, identifier, title, supported_formats,
                 data_format=None, abstract='', keywords=[], workdir=None, metadata=[], min_occurs=1,
                 max_occurs=1, mode=MODE.NONE,
                 default=None, default_type=basic.SOURCE_TYPE.DATA, translations=None):
        """constructor"""

        basic.ComplexInput.__init__(self, identifier, title=title,
                                    supported_formats=supported_formats,
                                    data_format=data_format, abstract=abstract,
                                    keywords=keywords, workdir=workdir, metadata=metadata,
                                    min_occurs=min_occurs,
                                    max_occurs=max_occurs, mode=mode,
                                    default=default, default_type=default_type, translations=translations)

        self.as_reference = False
        self.method = ''

    @property
    def json(self):
        """Get JSON representation of the input
        """
        data = {
            'identifier': self.identifier,
            'title': self.title,
            'abstract': self.abstract,
            'keywords': self.keywords,
            'metadata': [m.json for m in self.metadata],
            'type': 'complex',
            'data_format': self.data_format.json,
            'asreference': self.as_reference,
            'supported_formats': [frmt.json for frmt in self.supported_formats],
            'workdir': self.workdir,
            'mode': self.valid_mode,
            'min_occurs': self.min_occurs,
            'max_occurs': self.max_occurs,
            'translations': self.translations,
        }

        if self.prop == 'file':
            data['file'] = self.file
        elif self.prop == 'url':
            data["href"] = self.url
        elif self.prop == 'data':
            data = self._json_data(data)
        elif self.prop == 'stream':
            # we store the stream in the data property
            data = self._json_data(data)

        if self.data_format:
            if self.data_format.mime_type:
                data['mimetype'] = self.data_format.mime_type
            if self.data_format.encoding:
                data['encoding'] = self.data_format.encoding
            if self.data_format.schema:
                data['schema'] = self.data_format.schema

        return data

    @classmethod
    def from_json(cls, json_input):
        data_format = json_input.get('data_format')
        if data_format is not None:
            data_format = Format(
                schema=data_format.get('schema'),
                extension=data_format.get('extension'),
                mime_type=data_format.get('mime_type', ""),
                encoding=data_format.get('encoding')
            )
        instance = cls(
            identifier=json_input['identifier'],
            title=json_input.get('title'),
            abstract=json_input.get('abstract'),
            keywords=json_input.get('keywords', []),
            workdir=json_input.get('workdir'),
            metadata=[Metadata.from_json(data) for data in json_input.get('metadata', [])],
            data_format=data_format,
            supported_formats=[
                Format(
                    schema=infrmt.get('schema'),
                    extension=infrmt.get('extension'),
                    mime_type=infrmt.get('mime_type'),
                    encoding=infrmt.get('encoding')
                ) for infrmt in json_input.get('supported_formats', [])
            ],
            mode=json_input.get('mode', MODE.NONE),
            translations=json_input.get('translations'),
        )
        instance.as_reference = json_input.get('asreference', False)
        if json_input.get('file'):
            instance.file = json_input['file']
        elif json_input.get('href'):
            instance.url = json_input['href']
        elif json_input.get('data'):
            data = json_input['data']
            # remove cdata tag if it exists (issue #553)
            if isinstance(data, str):
                match = CDATA_PATTERN.match(data)
                if match:
                    data = match.group(1)
            instance.data = data

        return instance

    def _json_data(self, data):
        """Return Data node
        """

        if self.data:

            if self.data_format.mime_type in ["application/xml", "application/gml+xml", "text/xml"]:
                # Note that in a client-server round trip, the original and returned file will not be identical.
                data_doc = etree.parse(self.file)
                data["data"] = etree.tostring(data_doc, pretty_print=True).decode('utf-8')

            else:
                if self.data_format.encoding == 'base64':
                    data["data"] = self.base64.decode('utf-8')

                else:
                    # Otherwise we assume all other formats are unsafe and need to be enclosed in a CDATA tag.
                    if isinstance(self.data, bytes):
                        out = self.data.encode(self.data_format.encoding or 'utf-8')
                    else:
                        out = self.data

                    data["data"] = '<![CDATA[{}]]>'.format(out)

        return data

    def clone(self):
        """Create copy of yourself
        """
        return deepcopy(self)


class LiteralInput(basic.LiteralInput):
    """
    :param str identifier: The name of this input.
    :param str title: Title of the input
    :param pywps.inout.literaltypes.LITERAL_DATA_TYPES data_type: data type
    :param str workdir: working directory, to save temporary file objects in.
    :param str abstract: Input abstract
    :param list keywords: Keywords that characterize this input.
    :param list metadata: TODO
    :param str uoms: units
    :param int min_occurs: minimum occurence
    :param int max_occurs: maximum occurence
    :param pywps.validator.mode.MODE mode: validation mode (none to strict)
    :param pywps.inout.literaltypes.AnyValue allowed_values: or :py:class:`pywps.inout.literaltypes.AllowedValue` object
    :param metadata: List of metadata advertised by this process. They
                     should be :class:`pywps.app.Common.Metadata` objects.
    :param dict[str,dict[str,str]] translations: The first key is the RFC 4646 language code,
        and the nested mapping contains translated strings accessible by a string property.
        e.g. {"fr-CA": {"title": "Mon titre", "abstract": "Une description"}}
    """

    def __init__(self, identifier, title=None, data_type=None, workdir=None, abstract='', keywords=[],
                 metadata=[], uoms=None,
                 min_occurs=1, max_occurs=1,
                 mode=MODE.SIMPLE, allowed_values=None,
                 default=None, default_type=basic.SOURCE_TYPE.DATA, translations=None):
        """Constructor
        """
        data_type = data_type or 'string'
        basic.LiteralInput.__init__(self, identifier, title=title,
                                    data_type=data_type, workdir=workdir, abstract=abstract,
                                    keywords=keywords, metadata=metadata,
                                    uoms=uoms, min_occurs=min_occurs,
                                    max_occurs=max_occurs, mode=mode,
                                    allowed_values=allowed_values,
                                    default=default, default_type=default_type,
                                    translations=translations)

        self.as_reference = False

    @property
    def json(self):
        """Get JSON representation of the input
        """
        data = {
            'identifier': self.identifier,
            'title': self.title,
            'abstract': self.abstract,
            'keywords': self.keywords,
            'metadata': [m.json for m in self.metadata],
            'type': 'literal',
            'data_type': self.data_type,
            'workdir': self.workdir,
            'allowed_values': [value.json for value in self.allowed_values],
            'any_value': self.any_value,
            'mode': self.valid_mode,
            'min_occurs': self.min_occurs,
            'max_occurs': self.max_occurs,
            'translations': self.translations,
            # other values not set in the constructor
        }
        if self.values_reference:
            data['values_reference'] = self.values_reference.json
        if self.uoms:
            data["uoms"] = [uom.json for uom in self.uoms]
        if self.uom:
            data["uom"] = self.uom.json
        if self.data is not None:
            data['data'] = str(self.data)
        return data

    @classmethod
    def from_json(cls, json_input):
        allowed_values = []
        for allowed_value in json_input.get('allowed_values', []):
            if allowed_value['type'] == 'anyvalue':
                allowed_values.append(AnyValue())
            elif allowed_value['type'] == 'novalue':
                allowed_values.append(NoValue())
            elif allowed_value['type'] == 'valuesreference':
                allowed_values.append(ValuesReference.from_json(allowed_value))
            elif allowed_value['type'] == 'allowedvalue':
                allowed_values.append(AllowedValue.from_json(allowed_value))

        json_input_copy = deepcopy(json_input)
        json_input_copy['allowed_values'] = allowed_values
        json_input_copy['uoms'] = [
            basic.UOM(uom['uom'], uom['reference'])
            for uom in json_input.get('uoms', [])
        ]

        data = json_input_copy.pop('data', None)
        uom = json_input_copy.pop('uom', None)
        metadata = json_input_copy.pop('metadata', [])
        json_input_copy.pop('type', None)
        json_input_copy.pop('any_value', None)
        json_input_copy.pop('values_reference', None)

        instance = cls(**json_input_copy)

        instance.metadata = [Metadata.from_json(d) for d in metadata]
        instance.data = data
        if uom:
            instance.uom = basic.UOM(uom['uom'], uom['reference'])

        return instance

    def clone(self):
        """Create copy of yourself
        """
        return deepcopy(self)


def input_from_json(json_data):
    data_type = json_data.get('type', 'literal')
    if data_type in ['complex', 'reference']:
        inpt = ComplexInput.from_json(json_data)
    elif data_type == 'literal':
        inpt = LiteralInput.from_json(json_data)
    elif data_type == 'bbox':
        inpt = BoundingBoxInput.from_json(json_data)
    else:
        raise InvalidParameterValue("Input type not recognized: {}".format(data_type))

    return inpt



##################################################################
# Copyright 2018 Open Source Geospatial Foundation and others    #
# licensed under MIT, Please consult LICENSE.txt for details     #
##################################################################
import json
import tempfile
from pathlib import Path

import lxml
from pywps import xml_util as etree
import requests
from werkzeug.test import Client
from werkzeug.wrappers import Response
from pywps import __version__
from pywps import Process
from pywps.inout import LiteralInput, LiteralOutput, ComplexInput, ComplexOutput, BoundingBoxInput, BoundingBoxOutput
from pywps.inout import Format
from pywps.app.Common import Metadata, MetadataUrl

import re

import logging

logging.disable(logging.CRITICAL)


def service_ok(url, timeout=5):
    try:
        resp = requests.get(url, timeout=timeout)
        if 'html' in resp.headers['content-type']:
            ok = False
        else:
            ok = resp.ok
    except requests.exceptions.ReadTimeout:
        ok = False
    except requests.exceptions.ConnectTimeout:
        ok = False
    except Exception:
        ok = False
    return ok


class DocExampleProcess(Process):
    """This first line is going to be skipped by the :skiplines:1 option.

    Notes
    -----

    This is additional documentation that can be added following the Numpy docstring convention.
    """

    def __init__(self):
        inputs = [
            LiteralInput(
                'literal_input', "Literal input title", 'integer', abstract="Literal input value abstract.",
                min_occurs=0, max_occurs=1, uoms=['meters', 'feet'], default=1
            ),
            LiteralInput('date_input', 'The title is shown when no abstract is provided.', 'date',
                         allowed_values=['2000-01-01', '2018-01-01']),
            ComplexInput('complex_input', 'Complex input title',
                         [Format('application/json'), Format('application/x-netcdf')],
                         abstract="Complex input abstract.", ),
            BoundingBoxInput('bb_input', 'BoundingBox input title', ['EPSG:4326', ],
                             metadata=[Metadata('EPSG.io', 'http://epsg.io/'), ]),
        ]
        outputs = [
            LiteralOutput(
                'literal_output', 'Literal output title', 'boolean', abstract='Boolean output abstract.'
            ),
            ComplexOutput('complex_output', 'Complex output', [Format('text/plain'), ], ),
            BoundingBoxOutput('bb_output', 'BoundingBox output title', ['EPSG:4326', ])
        ]

        super(DocExampleProcess, self).__init__(
            self._handler,
            identifier='doc_example_process_identifier',
            title="Process title",
            abstract="Multiline process abstract.",
            version="4.0",
            metadata=[Metadata('PyWPS docs', 'https://pywps.org'),
                      Metadata('NumPy docstring conventions',
                               'https://github.com/numpy/numpy/blob/master/doc/HOWTO_DOCUMENT.rst.txt'),
                      MetadataUrl('Duplicate label', 'http://one.example.com', anonymous=True),
                      MetadataUrl('Duplicate label', 'http://two.example.com', anonymous=True),
                      ],
            inputs=inputs,
            outputs=outputs,
        )

    def _handler(self, request, response):
        pass


class WpsClient(Client):

    def post_xml(self, *args, **kwargs):
        doc = kwargs.pop('doc')
        data = etree.tostring(doc, pretty_print=True)
        kwargs['data'] = data
        return self.post(*args, **kwargs)

    def post_json(self, *args, **kwargs):
        doc = kwargs.pop('doc')
        # data = json.dumps(doc, indent=2)
        # kwargs['data'] = data
        kwargs['json'] = doc
        # kwargs['content_type'] = 'application/json'  # input is json, redundant as it's deducted from the json kwarg
        # kwargs['mimetype'] = 'application/json'  # output is json
        kwargs['environ_base'] = {'HTTP_ACCEPT': 'application/json'}  # output is json
        return self.post(*args, **kwargs)


class WpsTestResponse(Response):

    def __init__(self, *args):
        super(WpsTestResponse, self).__init__(*args)
        if re.match(r'text/xml(;\s*charset=.*)?', self.headers.get('Content-Type')):
            self.xml = etree.fromstring(self.get_data())

    def xpath(self, path):
        version = self.xml.attrib["version"]
        if version == "2.0.0":
            from pywps import namespaces200
            namespaces = namespaces200
        else:
            from pywps import namespaces100
            namespaces = namespaces100
        return self.xml.xpath(path, namespaces=namespaces)

    def xpath_text(self, path):
        return ' '.join(e.text for e in self.xpath(path))


def client_for(service):
    return WpsClient(service, WpsTestResponse)


def assert_response_accepted(resp):
    assert resp.status_code == 200
    assert re.match(r'text/xml(;\s*charset=.*)?', resp.headers['Content-Type'])
    success = resp.xpath_text('/wps:ExecuteResponse'
                              '/wps:Status'
                              '/wps:ProcessAccepted')
    assert success is not None
    # TODO: assert status URL is present


def assert_process_started(resp):
    assert resp.status_code == 200
    assert re.match(r'text/xml(;\s*charset=.*)?', resp.headers['Content-Type'])
    success = resp.xpath_text('/wps:ExecuteResponse'
                              '/wps:Status'
                              'ProcessStarted')
    # Is it still like this in PyWPS-4 ?
    assert success.split[0] == "processstarted"


def assert_response_success_json(resp, expected_data):
    assert resp.status_code == 200

    content_type = resp.headers['Content-Type']
    expected_contect_type = 'application/json'
    re_content_type = rf'{expected_contect_type}(;\s*charset=.*)?'
    assert re.match(re_content_type, content_type)

    data = json.loads(resp.data)

    success = data['status']['status']
    assert success == 'succeeded'

    if expected_data:
        outputs = data['outputs']
        assert outputs == expected_data


def assert_response_success(resp):
    assert resp.status_code == 200
    content_type = resp.headers['Content-Type']
    expected_contect_type = 'text/xml'
    re_content_type = rf'{expected_contect_type}(;\s*charset=.*)?'
    assert re.match(re_content_type, content_type)
    success = resp.xpath('/wps:ExecuteResponse/wps:Status/wps:ProcessSucceeded')
    assert len(success) == 1


def assert_process_exception(resp, code=None):
    assert resp.status_code == 400
    assert re.match(r'text/xml(;\s*charset=.*)?', resp.headers['Content-Type'])
    elem = resp.xpath('/ows:ExceptionReport'
                      '/ows:Exception')
    assert elem[0].attrib['exceptionCode'] == code


def assert_pywps_version(resp):
    # get first child of root element
    root_firstchild = resp.xpath('/*')[0].getprevious()
    assert isinstance(root_firstchild, lxml.etree._Comment)
    tokens = root_firstchild.text.split()
    assert len(tokens) == 2
    assert tokens[0] == 'PyWPS'
    assert tokens[1] == __version__


def assert_wps_version(response, version="1.0.0"):
    elem = response.xpath('/wps:Capabilities'
                          '/ows:ServiceIdentification'
                          '/ows:ServiceTypeVersion')
    found_version = elem[0].text
    assert version == found_version
    tmp = Path(tempfile.mkdtemp())
    with open(tmp / "out.xml", "wb") as out:
        out.writelines(response.response)



def get_translation(obj, attribute, language):
    """Get the translation from an object, for an attribute.

    The `obj` object is expected to have an attribute or key named `translations`
    and its value should be of type `dict[str,dict[str,str]]`.

    If the translation can't be found in the translations mapping,
    get the attribute on the object itself and raise
    :py:exc:`AttributeError` if it can't be found.

    The language property is converted to lowercase (see :py:func:`lower_case_dict`
    which must have been called on the translations first.

    :param str attribute: The attribute to get
    :param str language: The RFC 4646 language code
    """
    language = language.lower()
    try:
        return obj.translations[language][attribute]
    except (AttributeError, KeyError, TypeError):
        pass
    try:
        return obj["translations"][language][attribute]
    except (AttributeError, KeyError, TypeError):
        pass

    if hasattr(obj, attribute):
        return getattr(obj, attribute)

    try:
        return obj[attribute]
    except (TypeError, AttributeError):
        pass

    raise AttributeError(
        "Can't find translation '{}' for object type '{}'".format(attribute, type(obj).__name__)
    )


def lower_case_dict(translations=None):
    """Returns a new dict, with its keys converted to lowercase.

    :param dict[str, Any] translations: A dictionnary to be converted.
    """
    if translations is None:
        return
    return {k.lower(): v for k, v in translations.items()}



##################################################################
# Copyright 2018 Open Source Geospatial Foundation and others    #
# licensed under MIT, Please consult LICENSE.txt for details     #
##################################################################

from pywps.inout.inputs import LiteralInput, ComplexInput, BoundingBoxInput
from pywps.inout.outputs import LiteralOutput, ComplexOutput, BoundingBoxOutput
from pywps.inout.formats import Format, FORMATS, get_format
from pywps.inout.basic import UOM



##################################################################
# Copyright 2018 Open Source Geospatial Foundation and others    #
# licensed under MIT, Please consult LICENSE.txt for details     #
##################################################################

from pywps.processing.job import Job


class Processing(object):
    """
    :class:`Processing` is an interface for running jobs.
    """

    def __init__(self, process, wps_request, wps_response):
        self.job = Job(process, wps_request, wps_response)

    def start(self):
        raise NotImplementedError("Needs to be implemented in subclass.")

    def cancel(self):
        raise NotImplementedError("Needs to be implemented in subclass.")


class MultiProcessing(Processing):
    """
    :class:`MultiProcessing` is the default implementation to run jobs using the
    :module:`multiprocessing` module.
    """

    def start(self):
        import multiprocessing
        process = multiprocessing.Process(
            target=getattr(self.job.process, self.job.method),
            args=(self.job.wps_request, self.job.wps_response)
        )
        process.start()



##################################################################
# Copyright 2018 Open Source Geospatial Foundation and others    #
# licensed under MIT, Please consult LICENSE.txt for details     #
##################################################################

import logging
import os
from urllib.parse import urljoin
from pywps.exceptions import NotEnoughStorage, FileStorageError
from pywps import configuration as config
from pywps.inout.basic import IOHandler

from . import CachedStorage
from .implementationbuilder import StorageImplementationBuilder
from . import STORE_TYPE

LOGGER = logging.getLogger('PYWPS')


class FileStorageBuilder(StorageImplementationBuilder):

    def build(self):
        file_path = config.get_config_value('server', 'outputpath')
        base_url = config.get_config_value('server', 'outputurl')
        copy_function = config.get_config_value('server', 'storage_copy_function')
        return FileStorage(file_path, base_url, copy_function=copy_function)


def _build_output_name(output):
    (prefix, suffix) = os.path.splitext(output.file)
    if not suffix:
        suffix = output.data_format.extension
    _, file_name = os.path.split(prefix)
    output_name = file_name + suffix
    return (output_name, suffix)


class FileStorage(CachedStorage):
    """File storage implementation, stores data to file system

    >>> import ConfigParser
    >>> config = ConfigParser.RawConfigParser()
    >>> config.add_section('FileStorage')
    >>> config.set('FileStorage', 'target', './')
    >>> config.add_section('server')
    >>> config.set('server', 'outputurl', 'http://foo/bar/filestorage')
    >>>
    >>> store = FileStorage()
    >>>
    >>> class FakeOutput(object):
    ...     def __init__(self):
    ...         self.file = self._get_file()
    ...     def _get_file(self):
    ...         tiff_file = open('file.tiff', 'w')
    ...         tiff_file.close()
    ...         return 'file.tiff'
    >>> fake_out = FakeOutput()
    >>> (type, path, url) = store.store(fake_out)
    >>> type == STORE_TYPE.PATH
    True
    """

    def __init__(self, output_path, output_url, copy_function=None):
        """
        """
        CachedStorage.__init__(self)
        self.target = output_path
        self.output_url = output_url
        self.copy_function = copy_function

    def _do_store(self, output):
        import platform
        import math
        import tempfile
        import uuid

        file_name = output.file

        request_uuid = output.uuid or uuid.uuid1()

        # st.blksize is not available in windows, skips the validation on windows
        if platform.system() != 'Windows':
            file_block_size = os.stat(file_name).st_blksize
            # get_free_space delivers the numer of free blocks, not the available size!
            avail_size = get_free_space(self.target) * file_block_size
            file_size = os.stat(file_name).st_size

            # calculate space used according to block size
            actual_file_size = math.ceil(file_size / float(file_block_size)) * file_block_size

            if avail_size < actual_file_size:
                raise NotEnoughStorage('Not enough space in {} to store {}'.format(self.target, file_name))

        # create a target folder for each request
        target = os.path.join(self.target, str(request_uuid))
        if not os.path.exists(target):
            os.makedirs(target)

        # build output name
        output_name, suffix = _build_output_name(output)
        # build tempfile in case of duplicates
        if os.path.exists(os.path.join(target, output_name)):
            output_name = tempfile.mkstemp(suffix=suffix, prefix=file_name + '_',
                                           dir=target)[1]

        full_output_name = os.path.join(target, output_name)
        LOGGER.info(f'Storing file output to {full_output_name} ({self.copy_function}).')
        try:
            self.copy(output.file, full_output_name, self.copy_function)
        except Exception:
            LOGGER.exception(f"Could not copy {output_name}.")
            raise FileStorageError("Could not copy output file.")

        just_file_name = os.path.basename(output_name)

        url = self.url("{}/{}".format(request_uuid, just_file_name))
        LOGGER.info('File output URI: {}'.format(url))

        return (STORE_TYPE.PATH, output_name, url)

    @staticmethod
    def copy(src, dst, copy_function=None):
        """Copy file from source to destination using `copy_function`.

        Values of `copy_function` (default=`copy`):
        * copy: using `shutil.copy2`
        * move: using `shutil.move`
        * link: using `os.link`  (hardlink)
        """
        import shutil
        if copy_function == 'move':
            shutil.move(src, dst)
        elif copy_function == 'link':
            try:
                os.link(src, dst)
            except Exception:
                LOGGER.warn("Could not create hardlink. Fallback to copy.")
                FileStorage.copy(src, dst)
        else:
            shutil.copy2(src, dst)

    def write(self, data, destination, data_format=None):
        """
        Write data to self.target
        """
        if not os.path.exists(os.path.dirname(self.target)):
            os.makedirs(self.target)

        full_output_name = os.path.join(self.target, destination)

        with open(full_output_name, "w") as file:
            file.write(data)

        return self.url(destination)

    def url(self, destination):
        if isinstance(destination, IOHandler):
            output_name, _ = _build_output_name(destination)
            just_file_name = os.path.basename(output_name)
            dst = f"{destination.uuid}/{just_file_name}"
        else:
            dst = destination

        # make sure base url ends with '/'
        baseurl = self.output_url.rstrip('/') + '/'
        url = urljoin(baseurl, dst)
        return url

    def location(self, destination):
        return os.path.join(self.target, destination)


def get_free_space(folder):
    """ Return folder/drive free space (in bytes)
    """
    import platform

    if platform.system() == 'Windows':
        import ctypes

        free_bytes = ctypes.c_ulonglong(0)
        ctypes.windll.kernel32.GetDiskFreeSpaceExW(ctypes.c_wchar_p(folder), None, None, ctypes.pointer(free_bytes))
        free_space = free_bytes.value
    else:
        free_space = os.statvfs(folder).f_bfree

    LOGGER.debug('Free space: {}'.format(free_space))
    return free_space



##################################################################
# Copyright 2018 Open Source Geospatial Foundation and others    #
# licensed under MIT, Please consult LICENSE.txt for details     #
##################################################################

import json
import logging
import time
from werkzeug.wrappers import Request
from pywps import get_ElementMakerForVersion
from pywps.app.basic import get_response_type, get_json_indent, get_default_response_mimetype
from pywps.exceptions import NoApplicableCode
import pywps.configuration as config
from werkzeug.wrappers import Response

from pywps.inout.array_encode import ArrayEncoder
from pywps.response.status import WPS_STATUS
from pywps.response import WPSResponse
from pywps.inout.formats import FORMATS
from pywps.inout.outputs import ComplexOutput

import urllib.parse as urlparse
from urllib.parse import urlencode

LOGGER = logging.getLogger("PYWPS")

WPS, OWS = get_ElementMakerForVersion("1.0.0")


class ExecuteResponse(WPSResponse):

    def __init__(self, wps_request, uuid, **kwargs):
        """constructor

        :param pywps.app.WPSRequest.WPSRequest wps_request:
        :param pywps.app.Process.Process process:
        :param uuid: string this request uuid
        """

        super(ExecuteResponse, self).__init__(wps_request, uuid)

        self.process = kwargs["process"]
        self.outputs = {o.identifier: o for o in self.process.outputs}
        self.store_status_file = False

    # override WPSResponse._update_status
    def _update_status(self, status, message, status_percentage, clean=True):
        """
        Updates status report of currently running process instance:

        * Updates the status document.
        * Updates the status file (if requested).
        * Cleans the working directory when process has finished.

        This method is *only* called by pywps internally.
        """
        super(ExecuteResponse, self)._update_status(status, message, status_percentage)
        LOGGER.debug("_update_status: status={}, clean={}".format(status, clean))
        self._update_status_doc()
        if self.store_status_file:
            self._update_status_file()
        if clean:
            if self.status == WPS_STATUS.SUCCEEDED or self.status == WPS_STATUS.FAILED:
                LOGGER.debug("clean workdir: status={}".format(status))
                self.process.clean()

    def update_status(self, message, status_percentage=None):
        """
        Update status report of currently running process instance.

        This method is *only* called by the user provided process.
        The status is handled internally in pywps.

        :param str message: Message you need to share with the client
        :param int status_percentage: Percent done (number betwen <0-100>)
        """
        if status_percentage is None:
            status_percentage = self.status_percentage
        self._update_status(self.status, message, status_percentage, False)

    def _update_status_doc(self):
        try:
            # rebuild the doc
            self.doc, self.content_type = self._construct_doc()
        except Exception as e:
            raise NoApplicableCode('Building Response Document failed with : {}'.format(e))

    def _update_status_file(self):
        # TODO: check if file/directory is still present, maybe deleted in mean time
        try:
            # update the status xml file
            self.process.status_store.write(
                self.doc,
                self.process.status_filename,
                data_format=FORMATS.XML)
        except Exception as e:
            raise NoApplicableCode('Writing Response Document failed with : {}'.format(e))

    def _process_accepted(self):
        percent = int(self.status_percentage)
        if percent > 99:
            percent = 99
        return {
            "status": "accepted",
            "time": time.strftime('%Y-%m-%dT%H:%M:%SZ', time.localtime()),
            "percent_done": str(percent),
            "message": self.message
        }

    def _process_started(self):
        data = self._process_accepted()
        data.update({
            "status": "started",
        })
        return data

    def _process_paused(self):
        data = self._process_accepted()
        data.update({
            "status": "paused",
        })
        return data

    def _process_succeeded(self):
        data = self._process_accepted()
        data.update({
            "status": "succeeded",
            "percent_done": "100"
        })
        return data

    def _process_failed(self):
        data = self._process_accepted()
        data.update({
            "status": "failed",
            "code": "NoApplicableCode",
            "locator": "None",
        })
        return data

    def _get_serviceinstance(self):

        url = config.get_config_value("server", "url")
        params = {'request': 'GetCapabilities', 'service': 'WPS'}

        url_parts = list(urlparse.urlparse(url))
        query = dict(urlparse.parse_qsl(url_parts[4]))
        query.update(params)

        url_parts[4] = urlencode(query)
        return urlparse.urlunparse(url_parts).replace("&", "&amp;")

    @property
    def json(self):
        data = {}
        data["language"] = self.wps_request.language
        data["service_instance"] = self._get_serviceinstance()
        data["process"] = self.process.json

        if self.store_status_file:
            if self.process.status_location:
                data["status_location"] = self.process.status_url

        if self.status == WPS_STATUS.ACCEPTED:
            self.message = 'PyWPS Process {} accepted'.format(self.process.identifier)
            data["status"] = self._process_accepted()
        elif self.status == WPS_STATUS.STARTED:
            data["status"] = self._process_started()
        elif self.status == WPS_STATUS.FAILED:
            # check if process failed and display fail message
            data["status"] = self._process_failed()
        elif self.status == WPS_STATUS.PAUSED:
            # TODO: handle paused status
            data["status"] = self._process_paused()
        elif self.status == WPS_STATUS.SUCCEEDED:
            data["status"] = self._process_succeeded()
            # Process outputs XML
            data["outputs"] = [self.outputs[o].json for o in self.outputs]
        # lineage: add optional lineage when process has finished
        if self.status in [WPS_STATUS.SUCCEEDED, WPS_STATUS.FAILED]:
            # DataInputs and DataOutputs definition XML if lineage=true
            if self.wps_request.lineage == 'true':
                data["lineage"] = True
                try:
                    # TODO: stored process has ``pywps.inout.basic.LiteralInput``
                    # instead of a ``pywps.inout.inputs.LiteralInput``.
                    data["input_definitions"] = [self.wps_request.inputs[i][0].json for i in self.wps_request.inputs]
                except Exception as e:
                    LOGGER.error("Failed to update lineage for input parameter. {}".format(e))

                data["output_definitions"] = [self.outputs[o].json for o in self.outputs]
        return data

    @staticmethod
    def _render_json_response(jdoc):
        response = dict()
        response['status'] = jdoc['status']
        out = jdoc['process']['outputs']
        d = {}
        for val in out:
            id = val.get('identifier')
            if id is None:
                continue
            type = val.get('type')
            key = 'bbox' if type == 'bbox' else 'data'
            if key in val:
                d[id] = val[key]
        response['outputs'] = d
        return response

    def _construct_doc(self):
        if self.status == WPS_STATUS.SUCCEEDED and \
                hasattr(self.wps_request, 'preprocess_response') and \
                self.wps_request.preprocess_response:
            self.outputs = self.wps_request.preprocess_response(self.outputs,
                                                                request=self.wps_request,
                                                                http_request=self.wps_request.http_request)
        doc = self.json
        try:
            json_response, mimetype = get_response_type(
                self.wps_request.http_request.accept_mimetypes, self.wps_request.default_mimetype)
        except Exception:
            mimetype = get_default_response_mimetype()
            json_response = 'json' in mimetype
        if json_response:
            doc = json.dumps(self._render_json_response(doc), cls=ArrayEncoder, indent=get_json_indent())
        else:
            template = self.template_env.get_template(self.version + '/execute/main.xml')
            doc = template.render(**doc)
        return doc, mimetype

    @Request.application
    def __call__(self, request):
        accept_json_response, accepted_mimetype = get_response_type(
            self.wps_request.http_request.accept_mimetypes, self.wps_request.default_mimetype)
        if self.wps_request.raw:
            if self.status == WPS_STATUS.FAILED:
                return NoApplicableCode(self.message)
            else:
                wps_output_identifier = next(iter(self.wps_request.outputs))  # get the first key only
                wps_output_value = self.outputs[wps_output_identifier]
                response = wps_output_value.data
                if response is None:
                    return NoApplicableCode("Expected output was not generated")
                suffix = ''
                # if isinstance(wps_output_value, ComplexOutput):
                data_format = None
                if hasattr(wps_output_value, 'output_format'):
                    # this is set in the response, thus should be more precise
                    data_format = wps_output_value.output_format
                elif hasattr(wps_output_value, 'data_format'):
                    # this is set in the process' response _handler function, thus could have a few supported formats
                    data_format = wps_output_value.data_format
                if data_format is not None:
                    mimetype = data_format.mime_type
                    if data_format.extension is not None:
                        suffix = data_format.extension
                else:
                    # like LitearlOutput
                    mimetype = self.wps_request.outputs[wps_output_identifier].get('mimetype', None)
                if not isinstance(response, (str, bytes, bytearray)):
                    if not mimetype:
                        mimetype = accepted_mimetype
                    json_response = mimetype and 'json' in mimetype
                    if json_response:
                        mimetype = 'application/json'
                        suffix = '.json'
                        response = json.dumps(response, cls=ArrayEncoder, indent=get_json_indent())
                    else:
                        response = str(response)
                if not mimetype:
                    mimetype = None
                return Response(response, mimetype=mimetype,
                                headers={'Content-Disposition': 'attachment; filename="{}"'
                                         .format(wps_output_identifier + suffix)})
        else:
            if not self.doc:
                return NoApplicableCode("Output was not generated")
            return Response(self.doc, mimetype=accepted_mimetype)



from abc import abstractmethod
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from pywps import WPSRequest

from pywps.dblog import store_status
from pywps.response.status import WPS_STATUS
from pywps.translations import get_translation
from jinja2 import Environment, PackageLoader
import os


class RelEnvironment(Environment):
    """Override join_path() to enable relative template paths."""
    def join_path(self, template, parent):
        return os.path.dirname(parent) + '/' + template


def get_response(operation):

    from .capabilities import CapabilitiesResponse
    from .describe import DescribeResponse
    from .execute import ExecuteResponse

    if operation == "capabilities":
        return CapabilitiesResponse
    elif operation == "describe":
        return DescribeResponse
    elif operation == "execute":
        return ExecuteResponse


class WPSResponse(object):

    def __init__(self, wps_request: 'WPSRequest', uuid=None, version="1.0.0"):

        self.wps_request = wps_request
        self.uuid = uuid
        self.message = ''
        self.status = WPS_STATUS.ACCEPTED
        self.status_percentage = 0
        self.doc = None
        self.content_type = None
        self.version = version
        self.template_env = RelEnvironment(
            loader=PackageLoader('pywps', 'templates'),
            trim_blocks=True, lstrip_blocks=True,
            autoescape=True,
        )
        self.template_env.globals.update(get_translation=get_translation)

    def _update_status(self, status, message, status_percentage):
        """
        Update status report of currently running process instance

        :param str message: Message you need to share with the client
        :param int status_percentage: Percent done (number betwen <0-100>)
        :param pywps.response.status.WPS_STATUS status: process status - user should usually
            ommit this parameter
        """
        self.message = message
        self.status = status
        self.status_percentage = status_percentage
        store_status(self.uuid, self.status, self.message, self.status_percentage)

    @abstractmethod
    def _construct_doc(self):
        ...

    def get_response_doc(self):
        try:
            self.doc, self.content_type = self._construct_doc()
        except Exception as e:
            if hasattr(e, "description"):
                msg = e.description
            else:
                msg = e
            self._update_status(WPS_STATUS.FAILED, msg, 100)
            raise e

        else:
            self._update_status(WPS_STATUS.SUCCEEDED, "Response generated", 100)

            return self.doc, self.content_type



##################################################################
# Copyright 2018 Open Source Geospatial Foundation and others    #
# licensed under MIT, Please consult LICENSE.txt for details     #
##################################################################

import logging
import os
from pywps.translations import lower_case_dict
import sys
import traceback
import json
import shutil

from pywps import dblog
from pywps.response import get_response
from pywps.response.status import WPS_STATUS
from pywps.response.execute import ExecuteResponse
from pywps.app.WPSRequest import WPSRequest
from pywps.inout.inputs import input_from_json
from pywps.inout.outputs import output_from_json
import pywps.configuration as config
from pywps.exceptions import (StorageNotSupported, OperationNotSupported,
                              ServerBusy, NoApplicableCode,
                              InvalidParameterValue)
from pywps.app.exceptions import ProcessError
from pywps.inout.storage.builder import StorageBuilder
from pywps.inout.outputs import ComplexOutput
import importlib


LOGGER = logging.getLogger("PYWPS")


class Process(object):
    """
    :param handler: A callable that gets invoked for each incoming
                    request. It should accept a single
                    :class:`pywps.app.WPSRequest` argument and return a
                    :class:`pywps.app.WPSResponse` object.
    :param string identifier: Name of this process.
    :param string title: Human readable title of process.
    :param string abstract: Brief narrative description of the process.
    :param list keywords: Keywords that characterize a process.
    :param inputs: List of inputs accepted by this process. They
                   should be :class:`~LiteralInput` and :class:`~ComplexInput`
                   and :class:`~BoundingBoxInput`
                   objects.
    :param outputs: List of outputs returned by this process. They
                   should be :class:`~LiteralOutput` and :class:`~ComplexOutput`
                   and :class:`~BoundingBoxOutput`
                   objects.
    :param metadata: List of metadata advertised by this process. They
                     should be :class:`pywps.app.Common.Metadata` objects.
    :param dict[str,dict[str,str]] translations: The first key is the RFC 4646 language code,
        and the nested mapping contains translated strings accessible by a string property.
        e.g. {"fr-CA": {"title": "Mon titre", "abstract": "Une description"}}
    """

    def __init__(self, handler, identifier, title, abstract='', keywords=[], profile=[],
                 metadata=[], inputs=[], outputs=[], version='None', store_supported=False,
                 status_supported=False, grass_location=None, translations=None):
        self.identifier = identifier
        self.handler = handler
        self.title = title
        self.abstract = abstract
        self.keywords = keywords
        self.metadata = metadata
        self.profile = profile
        self.version = version
        self.inputs = inputs
        self.outputs = outputs
        self.uuid = None
        self._status_store = None
        # self.status_location = ''
        # self.status_url = ''
        self.workdir = None
        self._grass_mapset = None
        self.grass_location = grass_location
        self.service = None
        self.translations = lower_case_dict(translations)

        if store_supported:
            self.store_supported = 'true'
        else:
            self.store_supported = 'false'

        if status_supported:
            self.status_supported = 'true'
        else:
            self.status_supported = 'false'

    @property
    def json(self):

        return {
            'class': '{}:{}'.format(self.__module__, self.__class__.__name__),
            'uuid': str(self.uuid),
            'workdir': self.workdir,
            'version': self.version,
            'identifier': self.identifier,
            'title': self.title,
            'abstract': self.abstract,
            'keywords': self.keywords,
            'metadata': [m.json for m in self.metadata],
            'inputs': [i.json for i in self.inputs],
            'outputs': [o.json for o in self.outputs],
            'store_supported': self.store_supported,
            'status_supported': self.status_supported,
            'profile': [p for p in self.profile],
            'translations': self.translations,
        }

    @classmethod
    def from_json(cls, value):
        """init this process from json back again

        :param value: the json (not string) representation
        """
        module, classname = value['class'].split(':')
        # instantiate subclass of Process
        new_process = getattr(importlib.import_module(module), classname)()
        new_process._set_uuid(value['uuid'])
        new_process.set_workdir(value['workdir'])
        return new_process

    def execute(self, wps_request, uuid):
        self._set_uuid(uuid)
        self._setup_status_storage()
        self.async_ = False
        response_cls = get_response("execute")
        wps_response = response_cls(wps_request, process=self, uuid=self.uuid)

        LOGGER.debug('Check if status storage and updating are supported by this process')
        if wps_request.store_execute == 'true':
            if self.store_supported != 'true':
                raise StorageNotSupported('Process does not support the storing of the execute response')

            if wps_request.status == 'true':
                if self.status_supported != 'true':
                    raise OperationNotSupported('Process does not support the updating of status')

                wps_response.store_status_file = True
                self.async_ = True
            else:
                wps_response.store_status_file = False

        LOGGER.debug('Check if updating of status is not required then no need to spawn a process')

        wps_response = self._execute_process(self.async_, wps_request, wps_response)

        return wps_response

    def _set_uuid(self, uuid):
        """Set uuid and status location path and url
        """

        self.uuid = uuid
        for inpt in self.inputs:
            inpt.uuid = uuid

        for outpt in self.outputs:
            outpt.uuid = uuid

    def _setup_status_storage(self):
        self._status_store = StorageBuilder.buildStorage()

    @property
    def status_store(self):
        if self._status_store is None:
            self._setup_status_storage()
        return self._status_store

    @property
    def status_location(self):
        return self.status_store.location(self.status_filename)

    @property
    def status_filename(self):
        return str(self.uuid) + '.xml'

    @property
    def status_url(self):
        return self.status_store.url(self.status_filename)

    def _execute_process(self, async_, wps_request, wps_response):
        """Uses :module:`pywps.processing` module for sending process to
        background BUT first, check for maxprocesses configuration value

        :param async_: run in asynchronous mode
        :return: wps_response or None
        """

        maxparallel = int(config.get_config_value('server', 'parallelprocesses'))

        running, stored = dblog.get_process_counts()

        # async
        if async_:

            # run immedietly
            LOGGER.debug("Running processes: {} of {} allowed parallelprocesses".format(running, maxparallel))
            LOGGER.debug("Stored processes: {}".format(stored))

            if running < maxparallel or maxparallel == -1:
                wps_response._update_status(WPS_STATUS.ACCEPTED, "PyWPS Request accepted", 0)
                LOGGER.debug("Accepted request {}".format(self.uuid))
                self._run_async(wps_request, wps_response)

            # try to store for later usage
            else:
                maxprocesses = int(config.get_config_value('server', 'maxprocesses'))
                if stored >= maxprocesses and maxprocesses != -1:
                    raise ServerBusy('Maximum number of processes in queue reached. Please try later.')
                LOGGER.debug("Store process in job queue, uuid={}".format(self.uuid))
                dblog.store_process(self.uuid, wps_request)
                wps_response._update_status(WPS_STATUS.ACCEPTED, 'PyWPS Process stored in job queue', 0)

        # not async
        else:
            if running >= maxparallel and maxparallel != -1:
                raise ServerBusy('Maximum number of parallel running processes reached. Please try later.')
            wps_response._update_status(WPS_STATUS.ACCEPTED, "PyWPS Request accepted", 0)
            wps_response = self._run_process(wps_request, wps_response)

        return wps_response

    # This function may not raise exception and must return a valid wps_response
    # Failure must be reported as wps_response.status = WPS_STATUS.FAILED
    def _run_async(self, wps_request, wps_response):
        import pywps.processing
        process = pywps.processing.Process(
            process=self,
            wps_request=wps_request,
            wps_response=wps_response)
        LOGGER.debug("Starting process for request: {}".format(self.uuid))
        process.start()

    # This function may not raise exception and must return a valid wps_response
    # Failure must be reported as wps_response.status = WPS_STATUS.FAILED
    def _run_process(self, wps_request, wps_response):
        LOGGER.debug("Started processing request: {}".format(self.uuid))
        try:
            self._set_grass(wps_request)
            # if required set HOME to the current working directory.
            if config.get_config_value('server', 'sethomedir') is True:
                os.environ['HOME'] = self.workdir
                LOGGER.info('Setting HOME to current working directory: {}'.format(os.environ['HOME']))
            LOGGER.debug('ProcessID={}, HOME={}'.format(self.uuid, os.environ.get('HOME')))
            wps_response._update_status(WPS_STATUS.STARTED, 'PyWPS Process started', 0)
            self.handler(wps_request, wps_response)  # the user must update the wps_response.
            # Ensure process termination
            if wps_response.status != WPS_STATUS.SUCCEEDED and wps_response.status != WPS_STATUS.FAILED:
                # if (not wps_response.status_percentage) or (wps_response.status_percentage != 100):
                LOGGER.debug('Updating process status to 100% if everything went correctly')
                wps_response._update_status(WPS_STATUS.SUCCEEDED, f'PyWPS Process {self.title} finished', 100)
        except Exception as e:
            traceback.print_exc()
            LOGGER.debug('Retrieving file and line number where exception occurred')
            exc_type, exc_obj, exc_tb = sys.exc_info()
            found = False
            while not found:
                # search for the _handler method
                m_name = exc_tb.tb_frame.f_code.co_name
                if m_name == '_handler':
                    found = True
                else:
                    if exc_tb.tb_next is not None:
                        exc_tb = exc_tb.tb_next
                    else:
                        # if not found then take the first
                        exc_tb = sys.exc_info()[2]
                        break
            fname = os.path.split(exc_tb.tb_frame.f_code.co_filename)[1]
            method_name = exc_tb.tb_frame.f_code.co_name

            # update the process status to display process failed

            msg = 'Process error: method={}.{}, line={}, msg={}'.format(fname, method_name, exc_tb.tb_lineno, e)
            LOGGER.error(msg)
            # In case of a ProcessError use the validated exception message.
            if isinstance(e, ProcessError):
                msg = "Process error: {}".format(e)
            # Only in debug mode we use the log message including the traceback ...
            elif config.get_config_value("logging", "level") != "DEBUG":
                # ... otherwise we use a sparse common error message.
                msg = 'Process failed, please check server error log'
            wps_response._update_status(WPS_STATUS.FAILED, msg, 100)

        finally:
            # The run of the next pending request if finished here, weather or not it successful
            self.launch_next_process()

        return wps_response

    def launch_next_process(self):
        """Look at the queue of async process, if the queue is not empty launch the next pending request.
        """
        try:
            LOGGER.debug("Checking for stored requests")

            stored_request = dblog.pop_first_stored()
            if not stored_request:
                LOGGER.debug("No stored request found")
                return

            (uuid, request_json) = (stored_request.uuid, stored_request.request)
            request_json = request_json.decode('utf-8')
            LOGGER.debug("Launching the stored request {}".format(str(uuid)))
            new_wps_request = WPSRequest()
            new_wps_request.json = json.loads(request_json)
            process_identifier = new_wps_request.identifier
            process = self.service.prepare_process_for_execution(process_identifier)
            process._set_uuid(uuid)
            process._setup_status_storage()
            process.async_ = True
            process.setup_outputs_from_wps_request(new_wps_request)
            new_wps_response = ExecuteResponse(new_wps_request, process=process, uuid=uuid)
            new_wps_response.store_status_file = True
            process._run_async(new_wps_request, new_wps_response)
        except Exception as e:
            LOGGER.exception("Could not run stored process. {}".format(e))

    def clean(self):
        """Clean the process working dir and other temporary files
        """
        if config.get_config_value('server', 'cleantempdir'):
            LOGGER.info("Removing temporary working directory: {}".format(self.workdir))
            try:
                if os.path.isdir(self.workdir):
                    shutil.rmtree(self.workdir)
                if self._grass_mapset and os.path.isdir(self._grass_mapset):
                    LOGGER.info("Removing temporary GRASS GIS mapset: {}".format(self._grass_mapset))
                    shutil.rmtree(self._grass_mapset)
            except Exception as err:
                LOGGER.error('Unable to remove directory: {}'.format(err))
        else:
            LOGGER.warning('Temporary working directory is not removed: {}'.format(self.workdir))

    def set_workdir(self, workdir):
        """Set working dir for all inputs and outputs

        this is the directory, where all the data are being stored to
        """

        self.workdir = workdir
        for inpt in self.inputs:
            inpt.workdir = workdir

        for outpt in self.outputs:
            outpt.workdir = workdir

    def _set_grass(self, wps_request):
        """Handle given grass_location parameter of the constructor

        location is either directory name, 'epsg:1234' form or a georeferenced
        file

        in the first case, new temporary mapset within the location will be
        created

        in the second case, location will be created in self.workdir

        the mapset should be deleted automatically using self.clean() method
        """
        if self.grass_location:

            import random
            import string
            from grass.script import core as grass
            from grass.script import setup as gsetup

            # HOME needs to be set - and that is usually not the case for httpd
            # server
            os.environ['HOME'] = self.workdir

            # GISRC envvariable needs to be set
            gisrc = open(os.path.join(self.workdir, 'GISRC'), 'w')
            gisrc.write("GISDBASE: {}\n".format(self.workdir))
            gisrc.write("GUI: txt\n")
            gisrc.close()
            os.environ['GISRC'] = gisrc.name

            new_loc_args = dict()
            mapset_name = 'pywps_ms_{}'.format(
                ''.join(random.sample(string.ascii_letters, 5)))

            if self.grass_location.startswith('complexinput:'):
                # create new location from a georeferenced file
                ref_file_parameter = self.grass_location.split(':')[1]
                ref_file = wps_request.inputs[ref_file_parameter][0].file
                new_loc_args.update({'filename': ref_file})
            elif self.grass_location.lower().startswith('epsg:'):
                # create new location from epsg code
                epsg = self.grass_location.lower().replace('epsg:', '')
                new_loc_args.update({'epsg': epsg})

            if new_loc_args:
                dbase = self.workdir
                location = str()
                while os.path.isdir(os.path.join(dbase, location)):
                    location = 'pywps_loc_{}'.format(
                        ''.join(random.sample(string.ascii_letters, 5)))

                gsetup.init(os.environ['GISBASE'], dbase,
                            location, 'PERMANENT')

                grass.create_location(dbase=dbase,
                                      location=location,
                                      **new_loc_args)
                LOGGER.debug('GRASS location based on {} created'.format(
                    list(new_loc_args.keys())[0]))
                grass.run_command('g.mapset',
                                  mapset=mapset_name,
                                  flags='c',
                                  dbase=dbase,
                                  location=location,
                                  quiet=True)

            # create temporary mapset within existing location
            elif os.path.isdir(self.grass_location):
                from grass.pygrass.gis import make_mapset

                LOGGER.debug('Temporary mapset will be created')

                dbase = os.path.dirname(self.grass_location)
                location = os.path.basename(self.grass_location)

                grass.run_command('g.gisenv', set="GISDBASE={}".format(dbase))
                grass.run_command('g.gisenv', set="LOCATION_NAME=%s" % location)

                while os.path.isdir(os.path.join(dbase, location, mapset_name)):
                    mapset_name = 'pywps_ms_{}'.format(
                        ''.join(random.sample(string.ascii_letters, 5)))
                make_mapset(mapset=mapset_name, location=location,
                            gisdbase=dbase)
                grass.run_command('g.gisenv', set="MAPSET=%s" % mapset_name)

            else:
                raise NoApplicableCode('Location does exists or does not seem '
                                       'to be in "EPSG:XXXX" form nor is it existing directory: {}'.format(location))

            # set _grass_mapset attribute - will be deleted once handler ends
            self._grass_mapset = mapset_name

            # final initialization
            LOGGER.debug('GRASS Mapset set to {}'.format(mapset_name))

            LOGGER.debug('GRASS environment initialised')
            LOGGER.debug('GISRC {}, GISBASE {}, GISDBASE {}, LOCATION {}, MAPSET {}'.format(
                         os.environ.get('GISRC'), os.environ.get('GISBASE'),
                         dbase, location, os.path.basename(mapset_name)))

    def setup_outputs_from_wps_request(self, wps_request):
        # set as_reference to True for all the outputs specified as reference
        # if the output is not required to be raw
        if not wps_request.raw:
            for wps_outpt in wps_request.outputs:

                is_reference = wps_request.outputs[wps_outpt].get('asReference', 'false')
                mimetype = wps_request.outputs[wps_outpt].get('mimetype', '')
                if is_reference.lower() == 'true':
                    # check if store is supported
                    if self.store_supported == 'false':
                        raise StorageNotSupported(
                            'The storage of data is not supported for this process.')

                    is_reference = True
                else:
                    is_reference = False

                for outpt in self.outputs:
                    if outpt.identifier == wps_outpt:
                        outpt.as_reference = is_reference
                        if isinstance(outpt, ComplexOutput) and mimetype != '':
                            data_format = [f for f in outpt.supported_formats if f.mime_type == mimetype]
                            if len(data_format) == 0:
                                raise InvalidParameterValue(
                                    'MimeType ' + mimetype + ' not valid')
                            outpt.data_format = data_format[0]



##################################################################
# Copyright 2018 Open Source Geospatial Foundation and others    #
# licensed under MIT, Please consult LICENSE.txt for details     #
##################################################################

"""
Process exceptions raised intentionally in processes to provide information for users.
"""

import re

DEFAULT_ALLOWED_CHARS = ".:!?=,;-_/"

import logging

LOGGER = logging.getLogger('PYWPS')


def format_message(text, min_length=3, max_length=300, allowed_chars=None):
    allowed_chars = allowed_chars or DEFAULT_ALLOWED_CHARS
    special = re.escape(allowed_chars)
    pattern = rf'[\w{special}]+'
    msg = ' '.join(re.findall(pattern, text))
    msg.strip()
    if len(msg) >= min_length:
        msg = msg[:max_length]
    else:
        msg = ''
    return msg


class ProcessError(Exception):
    """:class:`pywps.app.exceptions.ProcessError` is an :class:`Exception`
    you can intentionally raise in a process
    to provide a user-friendly error message.
    The error message gets formatted (3<= message length <=300) and only
    alpha numeric characters and a few special characters are allowed.
    """
    default_msg = 'Sorry, process failed. Please check server error log.'

    def __init__(self, msg=None, min_length=3, max_length=300, allowed_chars=None):
        self.msg = msg
        self.min_length = min_length
        self.max_length = max_length
        self.allowed_chars = allowed_chars or DEFAULT_ALLOWED_CHARS

    def __str__(self):
        return self.message

    @property
    def message(self):
        try:
            msg = format_message(
                self.msg,
                min_length=self.min_length,
                max_length=self.max_length,
                allowed_chars=self.allowed_chars)
        except Exception as e:
            LOGGER.warning(f"process error formatting failed: {e}")
            msg = None
        if not msg:
            msg = self.default_msg
        return msg



import json

from werkzeug.wrappers import Request
import pywps.configuration as config
from pywps.app.basic import make_response, get_response_type, get_json_indent
from pywps.exceptions import NoApplicableCode
from pywps.exceptions import MissingParameterValue
from pywps.exceptions import InvalidParameterValue
from pywps.response import WPSResponse
from pywps import __version__
import os


class DescribeResponse(WPSResponse):

    def __init__(self, wps_request, uuid, **kwargs):

        super(DescribeResponse, self).__init__(wps_request, uuid)

        self.identifiers = None
        if "identifiers" in kwargs:
            self.identifiers = kwargs["identifiers"]
        self.processes = kwargs["processes"]

    @property
    def json(self):

        processes = []

        if 'all' in (ident.lower() for ident in self.identifiers):
            processes = (self.processes[p].json for p in self.processes)
        else:
            for identifier in self.identifiers:
                if identifier not in self.processes:
                    msg = "Unknown process {}".format(identifier)
                    raise InvalidParameterValue(msg, "identifier")
                else:
                    processes.append(self.processes[identifier].json)

        return {
            'pywps_version': __version__,
            'processes': processes,
            'language': self.wps_request.language,
        }

    @staticmethod
    def _render_json_response(jdoc):
        return jdoc

    def _construct_doc(self):
        if not self.identifiers:
            raise MissingParameterValue('Missing parameter value "identifier"', 'identifier')

        doc = self.json
        json_response, mimetype = get_response_type(
            self.wps_request.http_request.accept_mimetypes, self.wps_request.default_mimetype)
        if json_response:
            doc = json.dumps(self._render_json_response(doc), indent=get_json_indent())
        else:
            template = self.template_env.get_template(self.version + '/describe/main.xml')
            max_size = int(config.get_size_mb(config.get_config_value('server', 'maxsingleinputsize')))
            doc = template.render(max_size=max_size, **doc)
        return doc, mimetype

    @Request.application
    def __call__(self, request):
        # This function must return a valid response.
        try:
            doc, content_type = self.get_response_doc()
            return make_response(doc, content_type=content_type)
        except NoApplicableCode as e:
            return e
        except Exception as e:
            return NoApplicableCode(str(e))



##################################################################
# Copyright 2018 Open Source Geospatial Foundation and others    #
# licensed under MIT, Please consult LICENSE.txt for details     #
##################################################################
"""
XML tools
"""

import logging
from typing import Optional, Tuple

from werkzeug.wrappers import Response
import pywps.configuration as config

LOGGER = logging.getLogger('PYWPS')


def get_xpath_ns(version):
    """Get xpath namespace for specified WPS version
    currently 1.0.0 or 2.0.0 are supported
    """

    def xpath_ns(ele, path):
        """Function, which will return xpath namespace for given
        element and xpath
        """
        if version == "1.0.0":
            from pywps import namespaces100
            nsp = namespaces100
        elif version == "2.0.0":
            from pywps import namespaces200
            nsp = namespaces200
        return ele.xpath(path, namespaces=nsp)

    return xpath_ns


def make_response(doc, content_type):
    """response serializer"""
    if not content_type:
        content_type = get_default_response_mimetype()
    response = Response(doc, content_type=content_type)
    response.status_percentage = 100
    return response


def get_default_response_mimetype():
    default_mimetype = config.get_config_value('server', 'default_mimetype')
    return default_mimetype


def get_json_indent():
    json_ident = int(config.get_config_value('server', 'json_indent'))
    return json_ident if json_ident >= 0 else None


def get_response_type(accept_mimetypes, default_mimetype) -> Tuple[bool, str]:
    """
    This function determinate if the response should be JSON or XML based on
    the accepted mimetypes of the request and the default mimetype provided,
    which will be used in case both are equally accepted.

    :param accept_mimetypes: determinate which mimetypes are accepted
    :param default_mimetype: "text/xml", "application/json"
    :return: Tuple[bool, str] -
        bool - True: The response type is JSON, False: Otherwise - XML
        str - The output mimetype
    """
    accept_json = \
        accept_mimetypes.accept_json or \
        accept_mimetypes.best is None or \
        'json' in accept_mimetypes.best.lower()
    accept_xhtml = \
        accept_mimetypes.accept_xhtml or \
        accept_mimetypes.best is None or \
        'xml' in accept_mimetypes.best.lower()
    if not default_mimetype:
        default_mimetype = get_default_response_mimetype()
    json_is_default = 'json' in default_mimetype or '*' in default_mimetype
    json_response = (accept_json and (not accept_xhtml or json_is_default)) or \
                    (json_is_default and accept_json == accept_xhtml)
    mimetype = 'application/json' if json_response else 'text/xml' if accept_xhtml else ''
    return json_response, mimetype


def parse_http_url(http_request) -> dict:
    """
    This function parses the request URL and extracts the following:
        default operation, process identifier, output_ids, default mimetype
        info that cannot be terminated from the URL will be None (default)

        The url is expected to be in the following format, all the levels are optional.
        [base_url]/[identifier]/[output_ids]

    :param http_request: the request URL
    :return: dict with the extracted info listed:
        base_url - [wps|processes|jobs|api/api_level]
        default_mimetype - determinate by the base_url part:
            XML - if the base url == 'wps',
            JSON - if the base URL in ['api'|'jobs'|'processes']
        operation - also determinate by the base_url part:
            ['api'|'jobs'] -> 'execute'
            processes -> 'describeprocess' or 'getcapabilities'
                'describeprocess' if identifier is present as the next item, 'getcapabilities' otherwise
        api - api level, only expected if base_url=='api'
        identifier - the process identifier
        output_ids - if exist then it selects raw output with the name output_ids
    """
    operation = api = identifier = output_ids = default_mimetype = base_url = None
    if http_request:
        parts = str(http_request.path[1:]).split('/')
        i = 0
        if len(parts) > i:
            base_url = parts[i].lower()
            if base_url == 'wps':
                default_mimetype = 'xml'
            elif base_url in ['api', 'processes', 'jobs']:
                default_mimetype = 'json'
            i += 1
            if base_url == 'api':
                api = parts[i]
                i += 1
            if len(parts) > i:
                identifier = parts[i]
                i += 1
                if len(parts) > i:
                    output_ids = parts[i]
                    if not output_ids:
                        output_ids = None
    if base_url in ['jobs', 'api']:
        operation = 'execute'
    elif base_url == 'processes':
        operation = 'describeprocess' if identifier else 'getcapabilities'
    d = {}
    if operation:
        d['operation'] = operation
    if identifier:
        d['identifier'] = identifier
    if output_ids:
        d['output_ids'] = output_ids
    if default_mimetype:
        d['default_mimetype'] = default_mimetype
    if api:
        d['api'] = api
    if base_url:
        d['base_url'] = base_url
    return d



##################################################################
# Copyright 2018 Open Source Geospatial Foundation and others    #
# licensed under MIT, Please consult LICENSE.txt for details     #
##################################################################

import logging

LOGGER = logging.getLogger("PYWPS")


class Metadata(object):
    """
    ows:Metadata content model.

    :param title: Metadata title, human readable string
    :param href: fully qualified URL
    :param role: fully qualified URL
    :param type_: fully qualified URL
    """

    def __init__(self, title, href=None, role=None, type_='simple'):
        self.title = title
        self.href = href
        self.role = role
        self.type = type_

    def __iter__(self):
        metadata = {"title": self.title}

        if self.href is not None:
            metadata['href'] = self.href
        if self.role is not None:
            metadata['role'] = self.role
        metadata['type'] = self.type
        yield metadata

    @property
    def json(self):
        """Get JSON representation of the metadata
        """
        data = {
            'title': self.title,
            'href': self.href,
            'role': self.role,
            'type': self.type,
        }
        return data

    @classmethod
    def from_json(cls, json_input):
        instance = cls(
            title=json_input['title'],
            href=json_input['href'],
            role=json_input['role'],
            type_=json_input['type'],
        )
        return instance

    def __eq__(self, other):
        return all([
            self.title == other.title,
            self.href == other.href,
            self.role == other.role,
            self.type == other.type,
        ])


class MetadataUrl(Metadata):
    """Metadata subclass to allow anonymous links generation in documentation.

    Useful to avoid Sphinx "Duplicate explicit target name" warning.

    See https://docutils.sourceforge.io/docs/ref/rst/restructuredtext.html#anonymous-hyperlinks.

    Meant to use in documentation only, not needed in the xml response, nor being serialized or
    deserialized to/from json.  So that's why it is not directly in the base class.
    """

    def __init__(self, title, href=None, role=None, type_='simple',
                 anonymous=False):
        super().__init__(title, href=href, role=role, type_=type_)
        self.anonymous = anonymous
        "Whether to create anonymous link (boolean)."



##################################################################
# Copyright 2018 Open Source Geospatial Foundation and others    #
# licensed under MIT, Please consult LICENSE.txt for details     #
##################################################################

import platform
from typing import Union

from pathlib import Path
from urllib.parse import urlparse

is_windows = platform.system() == 'Windows'


def file_uri(path: Union[str, Path]) -> str:
    path = Path(path)
    path = path.as_uri()
    return str(path)


def uri_to_path(uri) -> str:
    p = urlparse(uri)
    path = p.path
    if is_windows:
        path = str(Path(path)).lstrip('\\')
    return path



##################################################################
# Copyright 2018 Open Source Geospatial Foundation and others    #
# licensed under MIT, Please consult LICENSE.txt for details     #
##################################################################
"""
WPS Output classes
"""
from typing import Optional, Sequence, Dict, Union

from pywps import xml_util as etree
import os
import re
from pywps.app.Common import Metadata
from pywps.exceptions import InvalidParameterValue
from pywps.inout import basic
from pywps.inout.storage.file import FileStorageBuilder
from pywps.inout.types import Translations
from pywps.validator.mode import MODE
from pywps import configuration as config
from pywps.inout.formats import Format, Supported_Formats, FORMATS


class BoundingBoxOutput(basic.BBoxOutput):
    """
    :param identifier: The name of this input.
    :param str title: Title of the input
    :param str abstract: Input abstract
    :param crss: List of supported coordinate reference system (e.g. ['EPSG:4326'])
    :param int dimensions: number of dimensions (2 or 3)
    :param int min_occurs: minimum occurence
    :param int max_occurs: maximum occurence
    :param pywps.validator.mode.MODE mode: validation mode (none to strict)
    :param metadata: List of metadata advertised by this process. They
                     should be :class:`pywps.app.Common.Metadata` objects.
    :param dict[str,dict[str,str]] translations: The first key is the RFC 4646 language code,
        and the nested mapping contains translated strings accessible by a string property.
        e.g. {"fr-CA": {"title": "Mon titre", "abstract": "Une description"}}
    """

    def __init__(self, identifier, title, crss, abstract='', keywords=[],
                 dimensions=2, metadata=[], min_occurs='1',
                 max_occurs='1', as_reference=False,
                 mode=MODE.NONE, translations=None):
        basic.BBoxOutput.__init__(self, identifier, title=title,
                                  abstract=abstract, keywords=keywords, crss=crss,
                                  dimensions=dimensions, mode=mode, translations=translations)

        self.metadata = metadata
        self.min_occurs = min_occurs
        self.max_occurs = max_occurs
        self.as_reference = as_reference

    @property
    def json(self):
        """Get JSON representation of the output
        """
        return {
            'identifier': self.identifier,
            'title': self.title,
            'abstract': self.abstract,
            'keywords': self.keywords,
            'min_occurs': self.min_occurs,
            'max_occurs': self.max_occurs,
            'metadata': self.metadata,
            'type': 'bbox',
            'crs': self.crs,
            'crss': self.crss,
            'dimensions': self.dimensions,
            'bbox': self.data,
            'll': self.ll,
            'ur': self.ur,
            'workdir': self.workdir,
            'mode': self.valid_mode,
            'translations': self.translations,
        }

    @classmethod
    def from_json(cls, json_output):
        instance = cls(
            identifier=json_output['identifier'],
            title=json_output['title'],
            abstract=json_output['abstract'],
            keywords=json_output['keywords'],
            min_occurs=json_output['min_occurs'],
            max_occurs=json_output['max_occurs'],
            metadata=[Metadata.from_json(data) for data in json_output.get('metadata', [])],
            crss=json_output['crss'],
            dimensions=json_output['dimensions'],
            mode=json_output['mode'],
            translations=json_output.get('translations'),
        )
        instance.data = json_output['bbox']
        instance.workdir = json_output['workdir']

        return instance


class ComplexOutput(basic.ComplexOutput):
    """
    :param identifier: The name of this output.
    :param title: Readable form of the output name.
    :param supported_formats: List of supported
        formats. The first format in the list will be used as the default.
    :type supported_formats: (pywps.inout.formats.Format, )
    :param str abstract: Description of the output
    :param pywps.validator.mode.MODE mode: validation mode (none to strict)
    :param metadata: List of metadata advertised by this process. They
                     should be :class:`pywps.app.Common.Metadata` objects.
    :param dict[str,dict[str,str]] translations: The first key is the RFC 4646 language code,
        and the nested mapping contains translated strings accessible by a string property.
        e.g. {"fr-CA": {"title": "Mon titre", "abstract": "Une description"}}
    """

    def __init__(self, identifier: str, title: str, supported_formats: Supported_Formats = None,
                 data_format=None, abstract: str = '', keywords=[], workdir=None,
                 metadata: Optional[Sequence[Metadata]] = None,
                 as_reference=False, mode: MODE = MODE.NONE, translations: Translations = None):
        if metadata is None:
            metadata = []

        basic.ComplexOutput.__init__(self, identifier, title=title,
                                     data_format=data_format, abstract=abstract, keywords=keywords,
                                     workdir=workdir,
                                     supported_formats=supported_formats,
                                     mode=mode, translations=translations)
        self.metadata = metadata
        self.as_reference = as_reference

        self.storage = None

    @property
    def json(self):
        """Get JSON representation of the output
        """
        data = {
            "identifier": self.identifier,
            "title": self.title,
            "abstract": self.abstract,
            'keywords': self.keywords,
            'type': 'complex',
            'supported_formats': [frmt.json for frmt in self.supported_formats],
            'asreference': self.as_reference,
            'data_format': self.data_format.json if self.data_format else None,
            'file': self.file if self.prop == 'file' else None,
            'workdir': self.workdir,
            'mode': self.valid_mode,
            'min_occurs': self.min_occurs,
            'max_occurs': self.max_occurs,
            'translations': self.translations,
        }

        if self.data_format:
            if self.data_format.mime_type:
                data['mimetype'] = self.data_format.mime_type
            if self.data_format.encoding:
                data['encoding'] = self.data_format.encoding
            if self.data_format.schema:
                data['schema'] = self.data_format.schema

        if self.as_reference:
            data = self._json_reference(data)
        else:
            data = self._json_data(data)

        return data

    @classmethod
    def from_json(cls, json_output):
        instance = cls(
            identifier=json_output['identifier'],
            title=json_output.get('title'),
            abstract=json_output.get('abstract'),
            keywords=json_output.get('keywords', []),
            workdir=json_output.get('workdir'),
            metadata=[Metadata.from_json(data) for data in json_output.get('metadata', [])],
            data_format=Format(
                schema=json_output['data_format'].get('schema'),
                extension=json_output['data_format'].get('extension'),
                mime_type=json_output['data_format']['mime_type'],
                encoding=json_output['data_format'].get('encoding')
            ),
            supported_formats=[
                Format(
                    schema=infrmt.get('schema'),
                    extension=infrmt.get('extension'),
                    mime_type=infrmt['mime_type'],
                    encoding=infrmt.get('encoding')
                ) for infrmt in json_output['supported_formats']
            ],
            mode=json_output.get('mode', MODE.NONE),
            translations=json_output.get('translations'),
        )
        instance.as_reference = json_output.get('asreference', False)
        if json_output.get('file'):
            instance.file = json_output['file']

        return instance

    def _json_reference(self, data):
        """Return Reference node
        """
        data["type"] = "reference"

        # get_url will create the file and return the url for it
        if self.prop == 'url':
            data["href"] = self.url
        elif self.prop is not None:
            self.storage = FileStorageBuilder().build()
            data["href"] = self.get_url()

        return data

    def _json_data(self, data):
        """Return Data node
        """
        data["type"] = "complex"

        if self.data:
            if self.data_format.mime_type in [FORMATS.GEOJSON.mime_type, FORMATS.JSON.mime_type]:
                data["data"] = self.data
            elif self.data_format.mime_type in ["application/xml", "application/gml+xml", "text/xml"]:
                # Note that in a client-server round trip, the original and returned file will not be identical.
                data_doc = etree.parse(self.file)
                data["data"] = etree.tostring(data_doc, pretty_print=True).decode('utf-8')

            else:
                if self.data_format.encoding == 'base64':
                    data["data"] = self.base64.decode('utf-8')
                else:
                    # Match only data that are safe CDATA pattern.
                    CDATA_PATTERN = re.compile(r'^<!\[CDATA\[((?!\]\]>).)*\]\]>$')

                    # Otherwise we assume all other formats are unsafe and need to be enclosed in a CDATA tag.
                    if isinstance(self.data, bytes):
                        # Try to inline data as text but if fail encode is in base64
                        if self.data_format.encoding == 'utf-8':
                            out = self.data.decode('utf-8')
                            # If data is already enclosed with CDATA pattern, do not add it twice
                            if CDATA_PATTERN.match(out):
                                data["data"] = out
                            else:
                                # Check if the data does not contain ]]> patern if is safe to use CDATA
                                # other wise we fallback to base64 encoding.
                                if not re.search('\\]\\]>', out):
                                    data["data"] = '<![CDATA[{}]]>'.format(out)
                                else:
                                    data['encoding'] = 'base64'  # override the unsafe encoding
                                    data["data"] = self.base64.decode('utf-8')
                        else:
                            data['encoding'] = 'base64'  # override the unsafe encoding
                            data["data"] = self.base64.decode('utf-8')
                    else:
                        out = str(self.data)
                        # If data is already enclose with CDATApatern do not add it twise
                        if CDATA_PATTERN.match(out):
                            data["data"] = out
                        else:
                            # Check if the data does not contain ]]> patern if is safe to use CDATA
                            # other wise we fallback to base64 encoding.
                            if not re.search('\\]\\]>', out):
                                data["data"] = '<![CDATA[{}]]>'.format(out)
                            else:
                                data['encoding'] = 'base64'  # override the unsafe encoding
                                data["data"] = self.base64.decode('utf-8')

        return data


class LiteralOutput(basic.LiteralOutput):
    """
    :param identifier: The name of this output.
    :param str title: Title of the input
    :param pywps.inout.literaltypes.LITERAL_DATA_TYPES data_type: data type
    :param str abstract: Input abstract
    :param str uoms: units
    :param pywps.validator.mode.MODE mode: validation mode (none to strict)
    :param metadata: List of metadata advertised by this process. They
                     should be :class:`pywps.app.Common.Metadata` objects.
    :param dict[str,dict[str,str]] translations: The first key is the RFC 4646 language code,
        and the nested mapping contains translated strings accessible by a string property.
        e.g. {"fr-CA": {"title": "Mon titre", "abstract": "Une description"}}
    """

    def __init__(self, identifier, title, data_type='string', abstract='', keywords=[],
                 metadata=[], uoms=None, mode=MODE.SIMPLE, translations=None):
        if uoms is None:
            uoms = []
        basic.LiteralOutput.__init__(self, identifier, title=title, abstract=abstract, keywords=keywords,
                                     data_type=data_type, uoms=uoms, mode=mode, translations=translations)
        self.metadata = metadata

    @property
    def json(self):
        """Get JSON representation of the output
        """
        data = {
            "identifier": self.identifier,
            "title": self.title,
            "abstract": self.abstract,
            "keywords": self.keywords,
            "data": self.data,
            "data_type": self.data_type,
            "type": "literal",
            "uoms": [u.json for u in self.uoms],
            "translations": self.translations,
        }

        if self.uom:
            data["uom"] = self.uom.json

        return data

    @classmethod
    def from_json(cls, json_output):
        uoms = [
            basic.UOM(uom['uom'], uom['reference'])
            for uom in json_output.get('uoms', [])
        ]
        uom = json_output.get('uom')

        instance = cls(
            identifier=json_output['identifier'],
            title=json_output['title'],
            data_type=json_output['data_type'],
            abstract=json_output['abstract'],
            keywords=json_output['keywords'],
            uoms=uoms,
            translations=json_output.get('translations'),
        )

        instance.data = json_output.get('data')
        if uom:
            instance.uom = basic.UOM(uom['uom'], uom['reference'])

        return instance


class MetaFile:
    """MetaFile object."""

    def __init__(self, identity=None, description=None, fmt=None):
        """Create a `MetaFile` object.

        :param str identity: human readable identity.
        :param str description: human readable file description.
        :param pywps.FORMAT fmt: file mime type.

        The content of each metafile is set like `ComplexOutputs`, ie
        using either the `data`, `file`, `stream` or `url` properties.

        The metalink document is created by a `MetaLink` instance, which
        holds a number of `MetaFile` instances.
        """
        self._size = None
        self._output = ComplexOutput(
            identifier=identity or '',
            title=description or '',
            as_reference=True,
            supported_formats=[fmt, ],
        )

    def _set_workdir(self, workdir):
        self._output.workdir = workdir

    @property
    def hash(self):
        """Text construct that conveys a cryptographic hash for a file.

        All hashes are encoded in lowercase hexadecimal format.  Hashes
        are used to verify the integrity of a complete file or portion
        of a file to determine if the file has been transferred without
        any errors.
        """
        import hashlib
        m = hashlib.sha256()
        with open(self.file, "rb") as f:
            for chunk in iter(lambda: f.read(4096), b""):
                m.update(chunk)
        return m.hexdigest()

    @property
    def identity(self):
        """Human-readable identity."""
        return self._output.identifier

    @property
    def name(self):
        """Indicate a specific file in a document describing multiple files."""
        js = self._output.json
        (_, name) = os.path.split(js.get('href', 'http:///'))
        return name

    @property
    def size(self):
        """Size of the linked content in bytes."""
        if self._size is None:
            self._size = self._output.size
        return self._size

    @size.setter
    def size(self, value):
        """Set size to avoid size calculation."""
        self._size = int(value)

    @property
    def urls(self):
        js = self._output.json
        return [js.get('href', ''), ]

    @property
    def mediatype(self):
        """Multipurpose Internet Mail Extensions (MIME) media type
        [RFC4288] of the metadata file available at the IRI."""
        return self._output.data_format.mime_type

    @property
    def data(self):
        return self._output.data

    @data.setter
    def data(self, value):
        self._output.data = value

    @property
    def file(self):
        return self._output.file

    @file.setter
    def file(self, value):
        self._output.file = value

    @property
    def url(self):
        return self._output.url

    @url.setter
    def url(self, value):
        self._output.url = value

    @property
    def stream(self):
        return self._output.stream

    @stream.setter
    def stream(self, value):
        self._output.stream = value

    def __str__(self):
        out = "MetaFile {}:".format(self.name)
        for url in self.urls:
            out += "\n\t{}".format(url)
        return out

    def __repr__(self):
        return "<pywps.inout.outputs.MetaFile {}>".format(self.name)


class MetaLink:
    _xml_template = 'metalink/3.0/main.xml'
    # Specs: https://www.metalinker.org/Metalink_3.0_Spec.pdf

    def __init__(self, identity=None, description=None, publisher=None, files=(),
                 workdir=None, checksums=False):
        """Create a MetaLink v3.0 instance.

        :param str identity: human readable identity.
        :param str description: human readable file description.
        :param str publisher: The name of the file's publisher.
        :param tuple files: Sequence of files to include in Metalink. Can also be added using `append`.
        :param str workdir: Work directory to store temporary files.
        :param bool checksums: Whether to compute checksums on files.

        To use, first append `MetaFile` instances, then write the metalink using the `xml`
        property.

        Methods:
            - `append`: add a `MetaFile` instance
        """
        self.identity = identity
        self.description = description
        self.workdir = workdir
        self.publisher = publisher
        self.files = []
        self.checksums = checksums

        for file in files:
            self.append(file)
        self._load_template()

    def append(self, file):
        """Append a `MetaFile` instance."""
        if not isinstance(file, MetaFile):
            raise ValueError("file must be a MetaFile instance.")
        file._set_workdir(self.workdir)
        self.files.append(file)

    @property
    def xml(self):
        return self._template.render(meta=self)

    @property
    def origin(self):
        """IRI where the Metalink Document was originally published.
        If the dynamic attribute of metalink:origin is "true", then
        updated versions of the Metalink can be found at this IRI.
        """
        return ""

    @property
    def published(self):
        """Date construct indicating an instant in time associated
        with an event early in the life cycle of the entry."""
        import datetime
        return datetime.datetime.now().strftime("%Y-%m-%dT%H:%M:%SZ")

    @property
    def generator(self):
        """Generating agent name and version used to generate a
        Metalink Document, for debugging and other purposes."""
        import pywps
        return "PyWPS/{}".format(pywps.__version__)

    @property
    def url(self):
        """Return the server URL."""
        return config.get_config_value('server', 'url')

    def _load_template(self):
        from pywps.response import RelEnvironment
        from jinja2 import PackageLoader

        template_env = RelEnvironment(
            loader=PackageLoader('pywps', 'templates'),
            trim_blocks=True, lstrip_blocks=True,
            autoescape=True, )

        self._template = template_env.get_template(self._xml_template)


class MetaLink4(MetaLink):
    _xml_template = 'metalink/4.0/main.xml'
    # Specs: https://tools.ietf.org/html/rfc5854


def output_from_json(json_data):
    data_type = json_data.get('type', 'literal')
    if data_type == 'complex':
        output = ComplexOutput.from_json(json_data)
    elif data_type == 'literal':
        output = LiteralOutput.from_json(json_data)
    elif data_type == 'bbox':
        output = BoundingBoxOutput.from_json(json_data)
    else:
        raise InvalidParameterValue("Output type not recognized: {}".format(data_type))

    return output
